""" SandboxMetadataDB class is a front-end to the metadata for sandboxes
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities import List
from DIRAC.Core.Security import Properties
from DIRAC.ConfigurationSystem.Client.Helpers import Registry


class SandboxMetadataDB(DB):
    def __init__(self):
        DB.__init__(self, "SandboxMetadataDB", "WorkloadManagement/SandboxMetadataDB")
        result = self.__initializeDB()
        if not result["OK"]:
            raise RuntimeError("Can't create tables: %s" % result["Message"])
        self.__assignedSBGraceDays = 0
        self.__unassignedSBGraceDays = 15

    def __initializeDB(self):
        """
        Create the tables
        """
        result = self._query("show tables")
        if not result["OK"]:
            return result

        tablesInDB = [t[0] for t in result["Value"]]
        tablesToCreate = {}
        self.__tablesDesc = {}

        self.__tablesDesc["sb_Owners"] = {
            "Fields": {
                "OwnerId": "INTEGER(10) UNSIGNED AUTO_INCREMENT NOT NULL",
                "Owner": "VARCHAR(32) NOT NULL",
                "OwnerDN": "VARCHAR(255) NOT NULL",
                "OwnerGroup": "VARCHAR(32) NOT NULL",
            },
            "PrimaryKey": "OwnerId",
        }

        self.__tablesDesc["sb_SandBoxes"] = {
            "Fields": {
                "SBId": "INTEGER(10) UNSIGNED AUTO_INCREMENT NOT NULL",
                "OwnerId": "INTEGER(10) UNSIGNED NOT NULL",
                "SEName": "VARCHAR(64) NOT NULL",
                "SEPFN": "VARCHAR(512) NOT NULL",
                "Bytes": "BIGINT(20) NOT NULL DEFAULT 0",
                "RegistrationTime": "DATETIME NOT NULL",
                "LastAccessTime": "DATETIME NOT NULL",
                "Assigned": "TINYINT NOT NULL DEFAULT 0",
            },
            "PrimaryKey": "SBId",
            "Indexes": {
                "SBOwner": ["OwnerId"],
            },
            "UniqueIndexes": {"Location": ["SEName", "SEPFN"]},
        }

        self.__tablesDesc["sb_EntityMapping"] = {
            "Fields": {
                "SBId": "INTEGER(10) UNSIGNED NOT NULL",
                "EntitySetup": "VARCHAR(64) NOT NULL",
                "EntityId": "VARCHAR(128) NOT NULL",
                "Type": "VARCHAR(64) NOT NULL",
            },
            "Indexes": {"Entity": ["EntityId", "EntitySetup"], "SBIndex": ["SBId"]},
            "UniqueIndexes": {"Mapping": ["SBId", "EntitySetup", "EntityId", "Type"]},
        }

        for tableName in self.__tablesDesc:
            if tableName not in tablesInDB:
                tablesToCreate[tableName] = self.__tablesDesc[tableName]

        return self._createTables(tablesToCreate)

    def registerAndGetOwnerId(self, owner, ownerDN, ownerGroup):
        """
        Get the owner ID and register it if it's not there
        """
        ownerEscaped = self._escapeString(owner)["Value"]
        ownerDNEscaped = self._escapeString(ownerDN)["Value"]
        ownerGroupEscaped = self._escapeString(ownerGroup)["Value"]
        sqlCmd = "SELECT OwnerId FROM `sb_Owners` WHERE Owner = %s AND OwnerDN = %s AND OwnerGroup = %s" % (
            ownerEscaped,
            ownerDNEscaped,
            ownerGroupEscaped,
        )
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        data = result["Value"]
        if data:
            return S_OK(data[0][0])
        # Its not there, insert it
        sqlCmd = "INSERT INTO `sb_Owners` ( OwnerId, Owner, OwnerDN, OwnerGroup ) VALUES ( 0, %s, %s, %s )" % (
            ownerEscaped,
            ownerDNEscaped,
            ownerGroupEscaped,
        )
        result = self._update(sqlCmd)
        if not result["OK"]:
            return result
        if "lastRowId" in result:
            return S_OK(result["lastRowId"])
        result = self._query("SELECT LAST_INSERT_ID()")
        if not result["OK"]:
            return S_ERROR("Can't determine owner id after insertion")
        return S_OK(result["Value"][0][0])

    def registerAndGetSandbox(self, owner, ownerDN, ownerGroup, sbSE, sbPFN, size=0):
        """
        Register a new sandbox in the metadata catalog
        Returns ( sbid, newSandbox )
        """
        result = self.registerAndGetOwnerId(owner, ownerDN, ownerGroup)
        if not result["OK"]:
            return result
        ownerId = result["Value"]
        sqlCmd = "INSERT INTO `sb_SandBoxes` ( SBId, OwnerId, SEName, SEPFN, Bytes, RegistrationTime, LastAccessTime )"
        sqlCmd = "%s VALUES ( 0, '%s', '%s', '%s', %d, UTC_TIMESTAMP(), UTC_TIMESTAMP() )" % (
            sqlCmd,
            ownerId,
            sbSE,
            sbPFN,
            size,
        )
        result = self._update(sqlCmd)
        if not result["OK"]:
            if result["Message"].find("Duplicate entry") == -1:
                return result
            # It's a duplicate, try to retrieve sbid
            sqlCond = ["SEPFN='%s'" % sbPFN, "SEName='%s'" % sbSE, "OwnerId='%s'" % ownerId]
            sqlCmd = "SELECT SBId FROM `sb_SandBoxes` WHERE %s" % " AND ".join(sqlCond)
            result = self._query(sqlCmd)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("Location %s already exists but doesn't belong to the user or setup")
            sbId = result["Value"][0][0]
            self.accessedSandboxById(sbId)
            return S_OK((sbId, False))
        # Inserted, time to get the id
        if "lastRowId" in result:
            return S_OK((result["lastRowId"], True))
        result = self._query("SELECT LAST_INSERT_ID()")
        if not result["OK"]:
            return S_ERROR("Can't determine sand box id after insertion")
        return S_OK((result["Value"][0][0], True))

    def accessedSandboxById(self, sbId):
        """
        Update last access time for sb id
        """
        return self.__accessedSandboxByCond({"SBId": sbId})

    def __accessedSandboxByCond(self, condDict):
        sqlCond = ["%s=%s" % (key, condDict[key]) for key in condDict]
        return self._update("UPDATE `sb_SandBoxes` SET LastAccessTime=UTC_TIMESTAMP() WHERE %s" % " AND ".join(sqlCond))

    def assignSandboxesToEntities(self, enDict, requesterName, requesterGroup, enSetup, ownerName="", ownerGroup=""):
        """
        Assign jobs to entities
        """

        if ownerName or ownerGroup:
            requesterProps = Registry.getPropertiesForEntity(requesterGroup, name=requesterName)
            if Properties.JOB_ADMINISTRATOR in requesterProps:
                if ownerName:
                    requesterName = ownerName
                if ownerGroup:
                    requesterGroup = ownerGroup

        entitiesToSandboxList = []
        for entityId in enDict:
            for sbTuple in enDict[entityId]:
                if not isinstance(sbTuple, (tuple, list)):
                    return S_ERROR("Entry for entity %s is not a itterable of tuples/lists" % entityId)
                if len(sbTuple) != 2:
                    return S_ERROR("SB definition is not ( SBLocation, Type )! It's '%s'" % str(sbTuple))
                SBLocation = sbTuple[0]
                if SBLocation.find("SB:") != 0:
                    return S_ERROR("%s doesn't seem to be a sandbox" % SBLocation)
                SBLocation = SBLocation[3:]
                splitted = List.fromChar(SBLocation, "|")
                if len(splitted) < 2:
                    return S_ERROR("SB Location has to have SEName|SEPFN form")
                SEName = splitted[0]
                SEPFN = ":".join(splitted[1:])
                entitiesToSandboxList.append((entityId, enSetup, sbTuple[1], SEName, SEPFN))
        if not entitiesToSandboxList:
            return S_OK()

        sbIds = []
        assigned = 0
        for entityId, entitySetup, SBType, SEName, SEPFN in entitiesToSandboxList:
            result = self.getSandboxId(SEName, SEPFN, requesterName, requesterGroup)
            insertValues = []
            if not result["OK"]:
                self.log.warn(
                    "Cannot find id for %s:%s with requester %s@%s" % (SEName, SEPFN, requesterName, requesterGroup)
                )
            else:
                sbId = result["Value"]
                sbIds.append(str(sbId))
                insertValues.append(
                    "( %s, %s, %s, %d )"
                    % (
                        self._escapeString(entityId)["Value"],
                        self._escapeString(entitySetup)["Value"],
                        self._escapeString(SBType)["Value"],
                        sbId,
                    )
                )

            if not insertValues:
                return S_ERROR(
                    "Sandbox does not exist or you're not authorized to assign it being %s@%s"
                    % (requesterName, requesterGroup)
                )
            sqlCmd = "INSERT INTO `sb_EntityMapping` ( entityId, entitySetup, Type, SBId ) VALUES %s" % ", ".join(
                insertValues
            )
            result = self._update(sqlCmd)
            if not result["OK"]:
                if result["Message"].find("Duplicate entry") == -1:
                    return result
            assigned += 1
        sqlCmd = "UPDATE `sb_SandBoxes` SET Assigned=1 WHERE SBId in ( %s )" % ", ".join(sbIds)
        result = self._update(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK(assigned)

    def __filterEntitiesByRequester(self, entitiesList, entitiesSetup, requesterName, requesterGroup):
        """
        Given a list of entities and a requester, return the ones that the requester is allowed to modify
        """
        sqlCond = ["s.OwnerId=o.OwnerId", "s.SBId=e.SBId", "e.EntitySetup=%s" % entitiesSetup]
        requesterProps = Registry.getPropertiesForEntity(requesterGroup, name=requesterName)
        if Properties.JOB_ADMINISTRATOR in requesterProps:
            # Do nothing, just ensure it doesn't fit in the other cases
            pass
        elif Properties.JOB_SHARING in requesterProps:
            sqlCond.append("o.OwnerGroup='%s'" % requesterGroup)
        elif Properties.NORMAL_USER in requesterProps:
            sqlCond.append("o.OwnerGroup='%s'" % requesterGroup)
            sqlCond.append("o.Owner='%s'" % requesterName)
        else:
            return S_ERROR("Not authorized to access sandbox")
        for i in range(len(entitiesList)):
            entitiesList[i] = self._escapeString(entitiesList[i])["Value"]
        if len(entitiesList) == 1:
            sqlCond.append("e.EntityId = %s" % entitiesList[0])
        else:
            sqlCond.append("e.EntityId in ( %s )" % ", ".join(entitiesList))
        sqlCmd = "SELECT DISTINCT e.EntityId FROM `sb_EntityMapping` e, `sb_SandBoxes` s, `sb_Owners` o WHERE"
        sqlCmd = "%s %s" % (sqlCmd, " AND ".join(sqlCond))
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK([row[0] for row in result["Value"]])

    def unassignEntities(self, entitiesDict, requesterName, requesterGroup):
        """
        Unassign jobs to sandboxes
        entitiesDict = { 'setup' : [ 'entityId', 'entityId' ] }
        """
        updated = 0
        for entitySetup in entitiesDict:
            entitiesIds = entitiesDict[entitySetup]
            if not entitiesIds:
                continue
            escapedSetup = self._escapeString(entitySetup)["Value"]
            result = self.__filterEntitiesByRequester(entitiesIds, escapedSetup, requesterName, requesterGroup)
            if not result["OK"]:
                gLogger.error("Cannot filter entities: %s" % result["Message"])
                continue
            ids = result["Value"]
            if not ids:
                return S_OK(0)
            sqlCond = ["EntitySetup = %s" % escapedSetup]
            sqlCond.append("EntityId in ( %s )" % ", ".join(["'%s'" % str(eid) for eid in ids]))
            sqlCmd = "DELETE FROM `sb_EntityMapping` WHERE %s" % " AND ".join(sqlCond)
            result = self._update(sqlCmd)
            if not result["OK"]:
                gLogger.error("Cannot unassign entities: %s" % result["Message"])
            else:
                updated += 1
        return S_OK(updated)

    def getSandboxesAssignedToEntity(self, entityId, entitySetup, requesterName, requesterGroup):
        """
        Get the sandboxes and the type of assignation to the jobId
        """
        sqlTables = ["`sb_SandBoxes` s", "`sb_EntityMapping` e"]
        sqlCond = [
            "s.SBId = e.SBId",
            "e.EntityId = %s" % self._escapeString(entityId)["Value"],
            "e.EntitySetup = %s" % self._escapeString(entitySetup)["Value"],
        ]
        requesterProps = Registry.getPropertiesForEntity(requesterGroup, name=requesterName)
        if Properties.JOB_ADMINISTRATOR in requesterProps or Properties.JOB_MONITOR in requesterProps:
            # Do nothing, just ensure it doesn't fit in the other cases
            pass
        elif Properties.JOB_SHARING in requesterProps:
            sqlTables.append("`sb_Owners` o")
            sqlCond.append("o.OwnerGroup='%s'" % requesterGroup)
            sqlCond.append("s.OwnerId=o.OwnerId")
        elif Properties.NORMAL_USER in requesterProps:
            sqlTables.append("`sb_Owners` o")
            sqlCond.append("o.OwnerGroup='%s'" % requesterGroup)
            sqlCond.append("o.Owner='%s'" % requesterName)
            sqlCond.append("s.OwnerId=o.OwnerId")
        else:
            return S_ERROR("Not authorized to access sandbox")
        sqlCmd = "SELECT DISTINCT s.SEName, s.SEPFN, e.Type FROM  %s WHERE %s" % (
            ", ".join(sqlTables),
            " AND ".join(sqlCond),
        )
        return self._query(sqlCmd)

    def getUnusedSandboxes(self):
        """
        Get sandboxes that have been assigned but the job is no longer there
        """
        sqlCond = [
            "Assigned AND SBId NOT IN ( SELECT SBId FROM `sb_EntityMapping` ) AND "
            "TIMESTAMPDIFF( DAY, LastAccessTime, UTC_TIMESTAMP() ) >= %d" % self.__assignedSBGraceDays,
            "! Assigned AND TIMESTAMPDIFF( DAY, LastAccessTime, UTC_TIMESTAMP() ) >= %s" % self.__unassignedSBGraceDays,
        ]
        sqlCmd = "SELECT SBId, SEName, SEPFN FROM `sb_SandBoxes` WHERE ( %s )" % " ) OR ( ".join(sqlCond)
        return self._query(sqlCmd)

    def deleteSandboxes(self, SBIdList):
        """
        Delete sandboxes
        """
        sqlSBList = ", ".join([str(sbid) for sbid in SBIdList])
        for table in ("sb_SandBoxes", "sb_EntityMapping"):
            sqlCmd = "DELETE FROM `%s` WHERE SBId IN ( %s )" % (table, sqlSBList)
            result = self._update(sqlCmd)
            if not result["OK"]:
                return result
        return S_OK()

    def getSandboxId(self, SEName, SEPFN, requesterName, requesterGroup, field="SBId", requesterDN=None):
        """
        Get the sandboxId if it exists

        :param SEName: name of the StorageElement
        :param SEPFN: PFN of the Sandbox
        :param requesterName: name (host or user) to use as credentials
        :param requesterGroup: user group used to use as credentials, or 'hosts'
        :param field: field we want to look for (default SBId)
        :param requestDN: host DN used as credentials

        :returns: S_OK with sandbox ID

        """
        sqlCond = [
            "s.SEPFN=%s" % self._escapeString(SEPFN)["Value"],
            "s.SEName=%s" % self._escapeString(SEName)["Value"],
            "s.OwnerId=o.OwnerId",
        ]
        sqlCmd = "SELECT s.%s FROM `sb_SandBoxes` s, `sb_Owners` o WHERE" % field
        requesterProps = Registry.getPropertiesForEntity(requesterGroup, name=requesterName, dn=requesterDN)
        if Properties.JOB_ADMINISTRATOR in requesterProps or Properties.JOB_MONITOR in requesterProps:
            # Do nothing, just ensure it doesn't fit in the other cases
            pass
        elif Properties.JOB_SHARING in requesterProps:
            sqlCond.append("o.OwnerGroup='%s'" % requesterGroup)
        elif Properties.NORMAL_USER in requesterProps:
            sqlCond.append("o.OwnerGroup='%s'" % requesterGroup)
            sqlCond.append("o.Owner='%s'" % requesterName)
        else:
            return S_ERROR("Not authorized to access sandbox")
        result = self._query("%s %s" % (sqlCmd, " AND ".join(sqlCond)))
        if not result["OK"]:
            return result
        data = result["Value"]
        if len(data) > 1:
            self.log.error("More than one sandbox registered with the same Id!", data)
        if not data:
            return S_ERROR("No sandbox matches the requirements")
        return S_OK(data[0][0])

    def getSandboxOwner(self, SEName, SEPFN, requesterDN, requesterGroup):
        """get the sandbox owner, if such sandbox exists

        :param SEName: name of the StorageElement
        :param SEPFN: PFN of the Sandbox
        :param requestDN: host DN used as credentials
        :param requesterGroup: group used to use as credentials (should be 'hosts')

        :returns: S_OK with tuple (owner, ownerDN, ownerGroup)
        """
        res = self.getSandboxId(SEName, SEPFN, None, requesterGroup, "OwnerId", requesterDN=requesterDN)
        if not res["OK"]:
            return res

        sqlCmd = "SELECT `Owner`, `OwnerDN`, `OwnerGroup` FROM `sb_Owners` WHERE `OwnerId` = %d" % res["Value"]
        res = self._query(sqlCmd)
        if not res["OK"]:
            return res
        return S_OK(res["Value"][0])



"""Collection of DIRAC useful file related modules.

.. warning::
   By default on Error they return None.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os
import hashlib
import random
import glob
import sys
import re
import errno

# Translation table of a given unit to Bytes
# I know, it should be kB...
SIZE_UNIT_CONVERSION = {
    "B": 1,
    "KB": 1024,
    "MB": 1024 * 1024,
    "GB": 1024 * 1024 * 1024,
    "TB": 1024 * 1024 * 1024 * 1024,
    "PB": 1024 * 1024 * 1024 * 1024 * 1024,
}


def mkDir(path):
    """Emulate 'mkdir -p path' (if path exists already, don't raise an exception)"""
    try:
        if os.path.isdir(path):
            return
        os.makedirs(path)
    except OSError as osError:
        if osError.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def mkLink(src, dst):
    """Protected creation of symbolic link"""
    try:
        os.symlink(src, dst)
    except OSError as osError:
        if osError.errno == errno.EEXIST and os.path.islink(dst) and os.path.realpath(dst) == src:
            pass
        else:
            raise


def makeGuid(fileName=None):
    """Utility to create GUID. If a filename is provided the
    GUID will correspond to its content's hexadecimal md5 checksum.
    Otherwise a random seed is used to create the GUID.
    The format is capitalized 8-4-4-4-12.

    .. warning::
       Could return None in case of OSError or IOError.

    :param string fileName: name of file
    """
    myMd5 = hashlib.md5()
    if fileName:
        try:
            with open(fileName, "rb") as fd:
                data = fd.read(10 * 1024 * 1024)
                myMd5.update(data)
        except Exception:
            return None
    else:
        myMd5.update(str(random.getrandbits(128)).encode())

    md5HexString = myMd5.hexdigest().upper()
    return generateGuid(md5HexString, "MD5")


def generateGuid(checksum, checksumtype):
    """Generate a GUID based on the file checksum"""

    if checksum:
        if checksumtype == "MD5":
            checksumString = checksum
        elif checksumtype == "Adler32":
            checksumString = str(checksum).zfill(32)
        else:
            checksumString = ""
        if checksumString:
            guid = "%s-%s-%s-%s-%s" % (
                checksumString[0:8],
                checksumString[8:12],
                checksumString[12:16],
                checksumString[16:20],
                checksumString[20:32],
            )
            guid = guid.upper()
            return guid

    # Failed to use the check sum, generate a new guid
    myMd5 = hashlib.md5()
    myMd5.update(str(random.getrandbits(128)).encode())
    md5HexString = myMd5.hexdigest()
    guid = "%s-%s-%s-%s-%s" % (
        md5HexString[0:8],
        md5HexString[8:12],
        md5HexString[12:16],
        md5HexString[16:20],
        md5HexString[20:32],
    )
    guid = guid.upper()
    return guid


def checkGuid(guid):
    """Checks whether a supplied GUID is of the correct format.
    The guid is a string of 36 characters [0-9A-F] long split into 5 parts of length 8-4-4-4-12.

    .. warning::
       As we are using GUID produced by various services and some of them could not follow
       convention, this function is passing by a guid which can be made of lower case chars or even just
       have 5 parts of proper length with whatever chars.

    :param string guid: string to be checked
    :return: True (False) if supplied string is (not) a valid GUID.
    """
    reGUID = re.compile("^[0-9A-F]{8}(-[0-9A-F]{4}){3}-[0-9A-F]{12}$")
    if reGUID.match(guid.upper()):
        return True
    else:
        guid = [len(x) for x in guid.split("-")]
        if guid == [8, 4, 4, 4, 12]:
            return True
    return False


def getSize(fileName):
    """Get size of a file.

    :param string fileName: name of file to be checked

    The os module claims only OSError can be thrown,
    but just for curiosity it's catching all possible exceptions.

    .. warning::
       On any exception it returns -1.

    """
    try:
        return os.stat(fileName)[6]
    except OSError:
        return -1


def getGlobbedTotalSize(files):
    """Get total size of a list of files or a single file.
    Globs the parameter to allow regular expressions.

    :params list files: list or tuple of strings of files
    """
    totalSize = 0
    if isinstance(files, (list, tuple)):
        for entry in files:
            size = getGlobbedTotalSize(entry)
            if size == -1:
                size = 0
            totalSize += size
    else:
        for path in glob.glob(files):
            if os.path.isdir(path) and not os.path.islink(path):
                for content in os.listdir(path):
                    totalSize += getGlobbedTotalSize(os.path.join(path, content))
            if os.path.isfile(path):
                size = getSize(path)
                if size == -1:
                    size = 0
                totalSize += size
    return totalSize


def getGlobbedFiles(files):
    """Get list of files or a single file.
    Globs the parameter to allow regular expressions.

    :params list files: list or tuple of strings of files
    """
    globbedFiles = []
    if isinstance(files, (list, tuple)):
        for entry in files:
            globbedFiles += getGlobbedFiles(entry)
    else:
        for path in glob.glob(files):
            if os.path.isdir(path) and not os.path.islink(path):
                for content in os.listdir(path):
                    globbedFiles += getGlobbedFiles(os.path.join(path, content))
            if os.path.isfile(path):
                globbedFiles.append(path)
    return globbedFiles


def getMD5ForFiles(fileList):
    """Calculate md5 for the content of all the files.

    :param fileList: list of paths
    :type fileList: python:list
    """
    fileList.sort()
    hashMD5 = hashlib.md5()
    for filePath in fileList:
        if os.path.isdir(filePath):
            continue
        with open(filePath, "rb") as fd:
            buf = fd.read(4096)
            while buf:
                hashMD5.update(buf)
                buf = fd.read(4096)
    return hashMD5.hexdigest()


def convertSizeUnits(size, srcUnit, dstUnit):
    """Converts a number from a given source unit to a destination unit.

    Example:
      In [1]: convertSizeUnits(1024, 'B', 'kB')
      Out[1]: 1

      In [2]: convertSizeUnits(1024, 'MB', 'kB')
      Out[2]: 1048576


    :param size: number to convert
    :param srcUnit: unit of the number. Any of ( 'B', 'kB', 'MB', 'GB', 'TB', 'PB')
    :param dstUnit: unit expected for the return. Any of ( 'B', 'kB', 'MB', 'GB', 'TB', 'PB')

    :returns: the size number converted in the dstUnit. In case of problem -sys.maxint is returned (negative)
    """

    srcUnit = srcUnit.upper()
    dstUnit = dstUnit.upper()

    try:
        convertedValue = float(size) * SIZE_UNIT_CONVERSION[srcUnit] / SIZE_UNIT_CONVERSION[dstUnit]
        return convertedValue

    # TypeError, ValueError: size is not a number
    # KeyError: srcUnit or dstUnit are not in the conversion list
    except (TypeError, ValueError, KeyError):
        return -sys.maxsize


if __name__ == "__main__":
    for p in sys.argv[1:]:
        print("%s : %s bytes" % (p, getGlobbedTotalSize(p)))



""" Module for handling AccountingDB tables on multiple DBs (e.g. 2 MySQL servers)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__RCSID__ = "$Id$"

from DIRAC import gConfig, S_OK, gLogger
from DIRAC.Core.Utilities.Plotting.TypeLoader import TypeLoader
from DIRAC.AccountingSystem.DB.AccountingDB import AccountingDB


class MultiAccountingDB(object):
    def __init__(self, csPath, readOnly=False):
        self.__csPath = csPath
        self.__readOnly = readOnly
        self.__dbByType = {}
        self.__defaultDB = "AccountingDB/AccountingDB"
        self.__log = gLogger.getSubLogger("MultiAccDB")
        self.__generateDBs()
        self.__registerMethods()

    def __generateDBs(self):
        self.__log.notice("Creating default AccountingDB...")
        self.__allDBs = {self.__defaultDB: AccountingDB(readOnly=self.__readOnly)}
        types = self.__allDBs[self.__defaultDB].getRegisteredTypes()
        result = gConfig.getOptionsDict(self.__csPath)
        if not result["OK"]:
            gLogger.verbose("No extra databases defined", "in %s" % self.__csPath)
            return
        validTypes = TypeLoader().getTypes()
        opts = result["Value"]
        for acType in opts:
            if acType not in validTypes:
                msg = "(%s defined in %s)" % (acType, self.__csPath)
                self.__log.fatal("Not a known accounting type", msg)
                raise RuntimeError(msg)
            dbName = opts[acType]
            gLogger.notice("Type will be assigned", "(%s to %s)" % (acType, dbName))
            if dbName not in self.__allDBs:
                fields = dbName.split("/")
                if len(fields) == 1:
                    dbName = "Accounting/%s" % dbName
                gLogger.notice("Creating DB", "%s" % dbName)
                self.__allDBs[dbName] = AccountingDB(dbName, readOnly=self.__readOnly)
            self.__dbByType[acType] = dbName

    def __registerMethods(self):
        for methodName in (
            "registerType",
            "changeBucketsLength",
            "regenerateBuckets",
            "deleteType",
            "insertRecordThroughQueue",
            "deleteRecord",
            "getKeyValues",
            "retrieveBucketedData",
            "calculateBuckets",
            "calculateBucketLengthForTime",
        ):
            (
                lambda closure: setattr(
                    self,
                    closure,
                    lambda *x: self.__mimeTypeMethod(closure, *x),  # pylint: disable=no-value-for-parameter
                )
            )(methodName)
        for methodName in (
            "autoCompactDB",
            "compactBuckets",
            "markAllPendingRecordsAsNotTaken",
            "loadPendingRecords",
            "getRegisteredTypes",
        ):
            (lambda closure: setattr(self, closure, lambda *x: self.__mimeMethod(closure, *x)))(methodName)

    def __mimeTypeMethod(self, methodName, setup, acType, *args):
        return getattr(self.__db(acType), methodName)("%s_%s" % (setup, acType), *args)

    def __mimeMethod(self, methodName, *args):
        end = S_OK()
        for dbName in self.__allDBs:
            res = getattr(self.__allDBs[dbName], methodName)(*args)
            if res and not res["OK"]:
                end = res
        return end

    def __db(self, acType):
        return self.__allDBs[self.__dbByType.get(acType, self.__defaultDB)]

    def insertRecordBundleThroughQueue(self, records):
        recByType = {}
        for record in records:
            acType = record[1]
            if acType not in recByType:
                recByType[acType] = []
            recByType[acType].append(("%s_%s" % (record[0], record[1]), record[2], record[3], record[4]))
        end = S_OK()
        for acType in recByType:
            res = self.__db(acType).insertRecordBundleThroughQueue(recByType[acType])
            if not res["OK"]:
                end = res
        return end



#!/usr/bin/env python
"""
Returns the LFN matching given GUIDs
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("GUIDs: GUIDs separated by a comma")
    Script.parseCommandLine()

    import DIRAC
    from DIRAC import gLogger

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    args = Script.getPositionalArgs()
    guids = args[0]

    try:
        guids = guids.split(",")
    except Exception:
        pass

    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    fc = FileCatalog()
    res = fc.getLFNForGUID(guids)
    if not res["OK"]:
        gLogger.error("Failed to get the LFNs", res["Message"])
        DIRAC.exit(-2)

    errorGuid = {}
    for guid, reason in res["Value"]["Failed"].items():
        errorGuid.setdefault(reason, []).append(guid)

    for error, guidList in errorGuid.items():
        gLogger.notice("Error '%s' for guids %s" % (error, guidList))

    for guid, lfn in res["Value"]["Successful"].items():
        gLogger.notice("%s -> %s" % (guid, lfn))

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" Base class for all the executor modules for Jobs Optimization
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import threading
import datetime  # Because eval(valenc) might require it
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import DEncode, List
from DIRAC.Core.Base.ExecutorModule import ExecutorModule
from DIRAC.WorkloadManagementSystem.Client.JobState.CachedJobState import CachedJobState
from DIRAC.WorkloadManagementSystem.Client import JobStatus, JobMinorStatus


class OptimizerExecutor(ExecutorModule):
    class JobLog:
        class LogWrap:
            def __init__(self, log, jid, funcName):
                self.__log = log
                self.__jid = jid
                self.__funcName = funcName

            def __call__(self, msg, varMsg=""):
                try:
                    funcObj = getattr(self.__log, self.__funcName)
                except AttributeError:
                    raise AttributeError("Logger does not have %s method" % self.__funcName)
                msg = "\n".join("[JID %s] %s" % (self.__jid, line) for line in msg.split("\n"))
                funcObj(msg, varMsg)

        def __init__(self, log, jid):
            self.__jid = jid
            self.__log = log

        def __bool__(self):
            return True

        # For Python 2 compatibility
        __nonzero__ = __bool__

        def __getattr__(self, name):
            return self.LogWrap(self.__log, self.__jid, name)

    @classmethod
    def initialize(cls):
        opName = cls.ex_getProperty("fullName")
        opName = "/".join(opName.split("/")[1:])
        if opName.find("Agent") == len(opName) - 5:
            opName = opName[:-5]
        cls.__optimizerName = opName
        maxTasks = cls.ex_getOption("Tasks", 1)
        cls.__jobData = threading.local()

        cls.__jobData.jobState = None
        cls.__jobData.jobLog = None
        cls.ex_setProperty("optimizerName", cls.__optimizerName)
        try:
            result = cls.initializeOptimizer()
            if not result["OK"]:
                return result
        except Exception as excp:
            cls.log.exception("Error while initializing optimizer")
            return S_ERROR("Error initializing: %s" % str(excp))
        cls.ex_setMind("WorkloadManagement/OptimizationMind")

        return S_OK()

    @classmethod
    def ex_optimizerName(cls):
        return cls.__optimizerName

    @classmethod
    def initializeOptimizer(cls):
        return S_OK()

    def processTask(self, jid, jobState):
        self.__jobData.jobState = jobState
        self.__jobData.jobLog = self.JobLog(self.log, jid)
        try:
            self.jobLog.info("Processing")
            optResult = self.optimizeJob(jid, jobState)
            # If the manifest is dirty, update it!
            result = jobState.getManifest()
            if not result["OK"]:
                self.jobLog.error("Failed to get job manifest", result["Message"])
                return result
            manifest = result["Value"]
            if manifest.isDirty():
                jobState.setManifest(manifest)
            # Did it go as expected? If not Failed!
            if not optResult["OK"]:
                self.jobLog.info("Set to Failed/%s" % optResult["Message"])
                minorStatus = "%s optimizer" % self.ex_optimizerName()
                return jobState.setStatus("Failed", minorStatus, optResult["Message"], source=self.ex_optimizerName())

            return S_OK()
        finally:
            self.__jobData.jobState = None
            self.__jobData.jobLog = None

    def optimizeJob(self, jid, jobState):
        raise Exception("You need to overwrite this method to optimize the job!")

    def setNextOptimizer(self, jobState=None):
        if not jobState:
            jobState = self.__jobData.jobState
        result = jobState.getOptParameter("OptimizerChain")
        if not result["OK"]:
            return result
        opChain = List.fromChar(result["Value"], ",")
        opName = self.__optimizerName
        try:
            opIndex = opChain.index(opName)
        except ValueError:
            return S_ERROR("Optimizer %s is not in the chain!" % opName)
        chainLength = len(opChain)
        if chainLength - 1 == opIndex:
            # This is the last optimizer in the chain!
            result = jobState.setStatus(
                JobStatus.WAITING, minorStatus=JobMinorStatus.PILOT_AGENT_SUBMISSION, appStatus="Unknown", source=opName
            )
            if not result["OK"]:
                return result

            result = jobState.insertIntoTQ()
            if not result["OK"]:
                return result

            return S_OK()
        # Keep optimizing!
        nextOp = opChain[opIndex + 1]
        self.jobLog.info("Set to Checking/%s" % nextOp)
        return jobState.setStatus(JobStatus.CHECKING, nextOp, source=opName)

    def storeOptimizerParam(self, name, value):
        if not self.__jobData.jobState:
            return S_ERROR("This function can only be called inside the optimizeJob function")
        valenc = DEncode.encode(value)
        return self.__jobData.jobState.setOptParameter(name, valenc)

    def retrieveOptimizerParam(self, name):
        if not self.__jobData.jobState:
            return S_ERROR("This function can only be called inside the optimizeJob function")
        result = self.__jobData.jobState.getOptParameter(name)
        if not result["OK"]:
            return result
        valenc = result["Value"]
        try:
            if not isinstance(valenc, bytes):
                valenc = valenc.encode()
            value, encLength = DEncode.decode(valenc)
            if encLength == len(valenc):
                return S_OK(value)
        except Exception:
            self.jobLog.exception("Opt param %s doesn't seem to be dencoded %r" % (name, valenc))
        return S_OK(eval(valenc))

    @property
    def jobLog(self):
        if not self.__jobData.jobLog:
            raise RuntimeError("jobLog can only be invoked inside the optimizeJob function")
        return self.__jobData.jobLog

    def deserializeTask(self, taskStub):
        return CachedJobState.deserialize(taskStub)

    def serializeTask(self, cjs):
        return S_OK(cjs.serialize())

    def fastTrackDispatch(self, jid, jobState):
        result = jobState.getStatus()
        if not result["OK"]:
            return S_ERROR("Could not retrieve job status for %s: %s" % (jid, result["Message"]))
        status, minorStatus = result["Value"]
        if status != JobStatus.CHECKING:
            self.log.info("[JID %s] Not in checking state. Avoid fast track" % jid)
            return S_OK()
        result = jobState.getOptParameter("OptimizerChain")
        if not result["OK"]:
            return S_ERROR("Could not retrieve OptimizerChain for job %s: %s" % (jid, result["Message"]))
        optChain = result["Value"]
        if minorStatus not in optChain:
            self.log.info("[JID %s] End of chain for job" % jid)
            return S_OK()
        self.log.info("[JID %s] Fast track possible to %s" % (jid, minorStatus))
        return S_OK("WorkloadManagement/%s" % minorStatus)



########################################################################
# File: FileCatalogProxyClient.py
########################################################################
""" File catalog client for the File Catalog proxy service """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.DISET.RPCClient import RPCClient


class FileCatalogProxyClient(object):
    """File catalog client for the File Catalog proxy service"""

    def __init__(self, fcName, **kwargs):
        """Constructor of the LCGFileCatalogProxy client class"""
        self.method = None
        self.fcName = fcName
        self.rpc = RPCClient("DataManagement/FileCatalogProxy", timeout=120)
        self.valid = False
        self.valid = self.rpc.ping()["OK"]
        self.interfaceMethods = None

    def isOK(self):
        """Is the Catalog available?"""
        return self.valid

    def getName(self):
        """Get the file catalog name"""
        return self.fcName

    def setInterfaceMethods(self, methodTuple):
        self.interfaceMethods = methodTuple

    def getInterfaceMethods(self):
        return self.interfaceMethods

    def __getattr__(self, name):
        self.method = name
        return self.execute

    def execute(self, *parms, **kws):
        """Magic method dispatcher"""
        return self.rpc.callProxyMethod(self.fcName, self.method, parms, kws)



"""
Utility functions for cloud endpoints.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys
import os
import six

from DIRAC import S_OK, S_ERROR
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

STATE_MAP = {
    0: "RUNNING",
    1: "REBOOTING",
    2: "TERMINATED",
    3: "PENDING",
    4: "UNKNOWN",
    5: "STOPPED",
    6: "SUSPENDED",
    7: "ERROR",
    8: "PAUSED",
}


def createMimeData(userDataTuple):

    userData = MIMEMultipart()
    for contents, mtype, fname in userDataTuple:
        try:
            mimeText = MIMEText(contents, mtype, sys.getdefaultencoding())
            mimeText.add_header("Content-Disposition", 'attachment; filename="%s"' % fname)
            userData.attach(mimeText)
        except Exception as e:
            return S_ERROR(str(e))

    return S_OK(userData)


def createPilotDataScript(vmParameters, bootstrapParameters):

    userDataDict = {}

    # Arguments to the vm-bootstrap command
    parameters = dict(vmParameters)
    parameters.update(bootstrapParameters)
    bootstrapArgs = {
        "dirac-site": parameters.get("Site"),
        "submit-pool": parameters.get("SubmitPool", ""),
        "ce-name": parameters.get("CEName"),
        "image-name": parameters.get("Image"),
        "vm-uuid": parameters.get("VMUUID"),
        "vmtype": parameters.get("VMType"),
        "vo": parameters.get("VO", ""),
        "running-pod": parameters.get("RunningPod", parameters.get("VO", "")),
        "cvmfs-proxy": parameters.get("CVMFSProxy", "DIRECT"),
        "cs-servers": ",".join(parameters.get("CSServers", [])),
        "number-of-processors": parameters.get("NumberOfProcessors", 1),
        "whole-node": parameters.get("WholeNode", True),
        "required-tag": parameters.get("RequiredTag", ""),
        "release-version": parameters.get("Version"),
        "lcgbundle-version": parameters.get("LCGBundleVersion", ""),
        "release-project": parameters.get("Project"),
        "setup": parameters.get("Setup"),
    }

    bootstrapString = ""
    for key, value in bootstrapArgs.items():
        bootstrapString += " --%s=%s \\\n" % (key, value)
    userDataDict["bootstrapArgs"] = bootstrapString

    userDataDict["user_data_commands_base_url"] = bootstrapParameters.get("user_data_commands_base_url")
    if not userDataDict["user_data_commands_base_url"]:
        return S_ERROR("user_data_commands_base_url is not defined")
    with open(bootstrapParameters["CloudPilotCert"]) as cfile:
        userDataDict["user_data_file_hostkey"] = cfile.read().strip()
    with open(bootstrapParameters["CloudPilotKey"]) as kfile:
        userDataDict["user_data_file_hostcert"] = kfile.read().strip()
    sshKey = None
    userDataDict["add_root_ssh_key"] = ""
    if "SshKey" in parameters:
        with open(parameters["SshKey"]) as sfile:
            sshKey = sfile.read().strip()
            userDataDict["add_root_ssh_key"] = (
                """
# Allow root login
sed -i 's/PermitRootLogin no/PermitRootLogin yes/g' /etc/ssh/sshd_config
# Copy id_rsa.pub to authorized_keys
echo \" """
                + sshKey
                + """\" > /root/.ssh/authorized_keys
service sshd restart
"""
            )

    # List of commands to be downloaded
    bootstrapCommands = bootstrapParameters.get("user_data_commands")
    if isinstance(bootstrapCommands, six.string_types):
        bootstrapCommands = bootstrapCommands.split(",")
    if not bootstrapCommands:
        return S_ERROR("user_data_commands list is not defined")
    userDataDict["bootstrapCommands"] = " ".join(bootstrapCommands)

    script = (
        """
cat <<X5_EOF >/root/hostkey.pem
%(user_data_file_hostkey)s
%(user_data_file_hostcert)s
X5_EOF
mkdir -p /var/spool/checkout/context
cd /var/spool/checkout/context
for dfile in %(bootstrapCommands)s
do
  echo curl --insecure -s %(user_data_commands_base_url)s/$dfile -o $dfile
  i=7
  while [ $i -eq 7 ]
  do
    curl --insecure -s %(user_data_commands_base_url)s/$dfile -o $dfile
    i=$?
    if [ $i -eq 7 ]; then
      echo curl connection failure for file $dfile
      sleep 10
    fi
  done
  curl --insecure -s %(user_data_commands_base_url)s/$dfile -o $dfile || echo Download of $dfile failed with $? !
done
%(add_root_ssh_key)s
chmod +x vm-bootstrap
/var/spool/checkout/context/vm-bootstrap %(bootstrapArgs)s
#/sbin/shutdown -h now
    """
        % userDataDict
    )

    if "HEPIX" in vmParameters:
        script = (
            """
cat <<EP_EOF >>/var/lib/hepix/context/epilog.sh
#!/bin/sh
%s
EP_EOF
chmod +x /var/lib/hepix/context/epilog.sh
      """
            % script
        )

    user_data = (
        """#!/bin/bash
mkdir -p /etc/joboutputs
(
%s
) > /etc/joboutputs/user_data.log 2>&1 &
exit 0
    """
        % script
    )

    cloud_config = """#cloud-config

output: {all: '| tee -a /var/log/cloud-init-output.log'}

cloud_final_modules:
  - [scripts-user, always]
    """
    # Also try to add ssh key using standart cloudinit approach(may not work)
    if sshKey:
        cloud_config += (
            """
users:
  - name: diracroot
    sudo: ALL=(ALL) NOPASSWD:ALL
    lock_passwd: true
    ssh-authorized-keys:
      - ssh-rsa %s
    """
            % sshKey
        )

    # print "AT >>> user_data", user_data
    # print "AT >>> cloud_config",  cloud_config

    return createMimeData(
        ((user_data, "x-shellscript", "dirac_boot.sh"), (cloud_config, "cloud-config", "cloud-config"))
    )


def createUserDataScript(parameters):

    defaultUser = os.environ.get("USER", parameters.get("User", "root"))
    sshUser = parameters.get("SshUser", defaultUser)
    defaultKey = os.path.expandvars("$HOME/.ssh/id_rsa.pub")
    sshKeyFile = parameters.get("SshKey", defaultKey)
    with open(sshKeyFile) as skf:
        sshKey = skf.read().strip()

    script = (
        """
# Allow root login
sed -i 's/PermitRootLogin no/PermitRootLogin yes/g' /etc/ssh/sshd_config
sed -i 's/PasswordAuthentication no/PasswordAuthentication yes/g' /etc/ssh/sshd_config
# Copy id_rsa.pub to authorized_keys
echo \" """
        + sshKey
        + """\" > /root/.ssh/authorized_keys
service sshd restart
"""
    )

    if "HEPIX" in parameters:
        script = (
            """
cat <<EP_EOF >>/var/lib/hepix/context/epilog.sh
#!/bin/sh
%s
EP_EOF
chmod +x /var/lib/hepix/context/epilog.sh
      """
            % script
        )

    user_data = (
        """#!/bin/bash
mkdir -p /etc/joboutputs
(
%s
) > /etc/joboutputs/user_data.log 2>&1 &
exit 0
    """
        % script
    )

    cloud_config = """#cloud-config

output: {all: '| tee -a /var/log/cloud-init-output.log'}

cloud_final_modules:
  - [scripts-user, always]
    """

    if sshKey:
        cloud_config += """
users:
  - name: %s
    sudo: ALL=(ALL) NOPASSWD:ALL
    lock_passwd: false
    ssh-authorized-keys:
      - %s
    """ % (
            sshUser,
            sshKey,
        )

        mime = createMimeData(
            ((user_data, "x-shellscript", "dirac_boot.sh"), (cloud_config, "cloud-config", "cloud-config"))
        )
        return mime


def createCloudInitScript(vmParameters, bootstrapParameters):
    """Create a user data script for cloud-init based images."""
    parameters = dict(vmParameters)
    parameters.update(bootstrapParameters)
    extraOpts = ""
    lcgVer = parameters.get("LCGBundleVersion", None)
    if lcgVer:
        extraOpts = "-g %s" % lcgVer

    # add extra yum installable packages
    extraPackages = ""
    if parameters.get("ExtraPackages"):
        packages = parameters.get("ExtraPackages")
        extraPackages = "\n".join([" - %s" % pp.strip() for pp in packages.split(",")])

    # add user account to connect by ssh
    sshUserConnect = ""
    sshUser = parameters.get("SshUser")
    sshKeyFile = parameters.get("SshKey")
    sshKey = ""
    if sshKeyFile:
        with open(sshKeyFile) as sshFile:
            sshKey = sshFile.read()
    if sshUser and sshKey:
        sshUserConnect = """
users:
  - name: %s
    sudo: ALL=(ALL) NOPASSWD:ALL
    lock_passwd: false
    ssh-authorized-keys:
      - %s
    """ % (
            sshUser,
            sshKey,
        )

    bootstrapArgs = {
        "dirac-site": parameters.get("Site"),
        "submit-pool": parameters.get("SubmitPool", ""),
        "ce-name": parameters.get("CEName"),
        "ce-type": parameters.get("InnerCEType", "Singularity"),
        "image-name": parameters.get("Image"),
        "vm-uuid": parameters.get("VMUUID"),
        "vmtype": parameters.get("VMType"),
        "vo": parameters.get("VO", ""),
        "running-pod": parameters.get("RunningPod", parameters.get("VO", "")),
        "cvmfs-proxy": parameters.get("CVMFSProxy", "DIRECT"),
        "cs-servers": ",".join(parameters.get("CSServers", [])),
        "number-of-processors": parameters.get("NumberOfProcessors", 1),
        "whole-node": parameters.get("WholeNode", True),
        "required-tag": parameters.get("RequiredTag", ""),
        "release-version": parameters.get("Version"),
        "extraopts": extraOpts,
        "release-project": parameters.get("Project"),
        "setup": parameters.get("Setup"),
        "user-root": parameters.get("UserRoot", "/cvmfs/cernvm-prod.cern.ch/cvm4"),
        "timezone": parameters.get("Timezone", "UTC"),
        "pilot-server": parameters.get("pilotFileServer", "localhost"),
        "extra-packages": extraPackages,
        "ssh-user": sshUserConnect,
        "max-cycles": parameters.get("MaxCycles", "100"),
    }
    default_template = os.path.join(os.path.dirname(__file__), "cloudinit.template")
    template_path = parameters.get("CITemplate", default_template)
    # Cert/Key need extra indents to keep yaml formatting happy
    with open(bootstrapParameters["CloudPilotCert"]) as cfile:
        raw_str = cfile.read().strip()
        raw_str = raw_str.replace("\n", "\n     ")
        bootstrapArgs["hostkey"] = raw_str
    with open(bootstrapParameters["CloudPilotKey"]) as kfile:
        raw_str = kfile.read().strip()
        raw_str = raw_str.replace("\n", "\n     ")
        bootstrapArgs["hostcert"] = raw_str
    with open(template_path) as template_fd:
        template = template_fd.read()
    template = template % bootstrapArgs
    mime = createMimeData(((template, "cloud-config", "pilotconfig"),))
    return mime



#!/bin/env python
"""
List the number of requests in the caches of all the ReqProxyies
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("", "Full", "   Print full list of requests")
    Script.parseCommandLine()
    from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient

    fullPrint = False

    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Full":
            fullPrint = True

    reqClient = ReqClient()

    for server, rpcClient in reqClient.requestProxies().items():
        DIRAC.gLogger.always("Checking request cache at %s" % server)
        reqCache = rpcClient.listCacheDir()
        if not reqCache["OK"]:
            DIRAC.gLogger.error("Cannot list request cache", reqCache)
            continue
        reqCache = reqCache["Value"]

        if not reqCache:
            DIRAC.gLogger.always("No request in cache")
        else:
            if fullPrint:
                DIRAC.gLogger.always("List of requests", reqCache)
            else:
                DIRAC.gLogger.always("Number of requests in the cache", len(reqCache))

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-dms-replicate-lfn
# Author  : Stuart Paterson
########################################################################
"""
Replicate an existing LFN to another Storage Element

Example:
  $ dirac-dms-replicate-lfn /formation/user/v/vhamar/Test.txt DIRAC-USER
  {'Failed': {},
   'Successful': {'/formation/user/v/vhamar/Test.txt': {'register': 0.50833415985107422,
                                                        'replicate': 11.878520965576172}}}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("LFN:      Logical File Name or file containing LFNs")
    Script.registerArgument("Dest:     Valid DIRAC SE")
    Script.registerArgument("Source:   Valid DIRAC SE", default="", mandatory=False)
    Script.registerArgument("Cache:    Local directory to be used as cache", default="", mandatory=False)
    _, args = Script.parseCommandLine(ignoreErrors=True)

    if len(args) > 4:
        Script.showHelp(exitCode=1)

    lfn, seName, sourceSE, localCache = Script.getPositionalArgs(group=True)

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    exitCode = 0

    try:
        f = open(lfn, "r")
        lfns = f.read().splitlines()
        f.close()
    except Exception:
        lfns = [lfn]

    finalResult = {"Failed": [], "Successful": []}
    for lfn in lfns:
        result = dirac.replicateFile(lfn, seName, sourceSE, localCache, printOutput=True)
        if not result["OK"]:
            finalResult["Failed"].append(lfn)
            print("ERROR %s" % (result["Message"]))
            exitCode = 2
        else:
            finalResult["Successful"].append(lfn)

    if len(lfns) > 1:
        print(finalResult)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Module containing a front-end to the ElasticSearch-based ElasticJobParametersDB.
    This module interacts with one ES index: "ElasticJobParametersDB",
    which is a drop-in replacement for MySQL-based table JobDB.JobParameters.
    While JobDB.JobParameters in MySQL is defined as::

      CREATE TABLE `JobParameters` (
        `JobID` INT(11) UNSIGNED NOT NULL,
        `Name` VARCHAR(100) NOT NULL,
        `Value` BLOB NOT NULL,
        PRIMARY KEY (`JobID`,`Name`),
        FOREIGN KEY (`JobID`) REFERENCES `Jobs`(`JobID`)
      ) ENGINE=InnoDB DEFAULT CHARSET=latin1;

    Here we define a mapping as::

      "JobID": {"type": "long"},
      "Name": {"type": "keyword"},
      "Value": {"type": "text"}

    which is an "equivalent" representation.

    The reason for switching to a ES-based JobParameters lies in the extended searching
    capabilities of ES (ES will analyze+index text fields, while MySQL won't do that on BLOB types).
    This results in higher traceability for DIRAC jobs.

    The following class methods are provided for public usage
      - getJobParameters()
      - setJobParameter()
      - deleteJobParameters()
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

__RCSID__ = "$Id$"

from DIRAC import S_OK, gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals
from DIRAC.Core.Base.ElasticDB import ElasticDB

name = "ElasticJobParametersDB"

mapping = {"properties": {"JobID": {"type": "long"}, "Name": {"type": "keyword"}, "Value": {"type": "text"}}}


class ElasticJobParametersDB(ElasticDB):
    def __init__(self):
        """Standard Constructor"""

        try:
            section = getDatabaseSection("WorkloadManagement/ElasticJobParametersDB")
            indexPrefix = gConfig.getValue("%s/IndexPrefix" % section, CSGlobals.getSetup()).lower()

            # Connecting to the ES cluster
            super(ElasticJobParametersDB, self).__init__(name, "WorkloadManagement/ElasticJobParametersDB", indexPrefix)
        except Exception as ex:
            self.log.error("Can't connect to ElasticJobParametersDB", repr(ex))
            raise RuntimeError("Can't connect to ElasticJobParametersDB")

        self.indexName = "%s_%s" % (self.getIndexPrefix(), name.lower())
        # Verifying if the index is there, and if not create it
        res = self.existingIndex(self.indexName)
        if not res["OK"] or not res["Value"]:
            result = self.createIndex(self.indexName, mapping, period=None)
            if not result["OK"]:
                self.log.error(result["Message"])
                raise RuntimeError(result["Message"])
            self.log.always("Index created:", self.indexName)

        self.dslSearch = self._Search(self.indexName)

    def getJobParameters(self, jobID, paramList=None):
        """Get Job Parameters defined for jobID.
          Returns a dictionary with the Job Parameters.
          If paramList is empty - all the parameters are returned.

        :param self: self reference
        :param int jobID: Job ID
        :param list paramList: list of parameters to be returned (also a string is treated)

        :return: dict with all Job Parameter values
        """

        if paramList:
            if isinstance(paramList, six.string_types):
                paramList = paramList.replace(" ", "").split(",")
        else:
            paramList = []

        self.log.debug("JobDB.getParameters: Getting Parameters for job %s" % jobID)

        resultDict = {}

        # the following should be equivalent to
        # {
        #   "query": {
        #     "bool": {
        #       "filter": {  # no scoring
        #         "term": {"JobID": jobID}  # term level query, does not pass through the analyzer
        #       }
        #     }
        #   }
        # }

        s = self.dslSearch.query("bool", filter=self._Q("term", JobID=jobID))

        res = s.execute()

        for hit in res:
            name = hit.Name
            if paramList and name not in paramList:
                continue
            resultDict[name] = hit.Value

        return S_OK({jobID: resultDict})

    def setJobParameter(self, jobID, key, value):
        """
        Inserts data into ElasticJobParametersDB index

        :param self: self reference
        :param int jobID: Job ID
        :param str key: parameter key
        :param str value: parameter value

        :returns: S_OK/S_ERROR as result of indexing
        """
        data = {"JobID": jobID, "Name": key, "Value": value}

        self.log.debug("Inserting data in %s:%s" % (self.indexName, data))

        result = self.index(self.indexName, body=data, docID=str(jobID) + key)
        if not result["OK"]:
            self.log.error("ERROR: Couldn't insert data", result["Message"])
        return result

    def setJobParameters(self, jobID, parameters):
        """
        Inserts data into ElasticJobParametersDB index using bulk indexing

        :param self: self reference
        :param int jobID: Job ID
        :param list parameters: list of tuples (name, value) pairs

        :returns: S_OK/S_ERROR as result of indexing
        """
        self.log.debug("Inserting parameters", "in %s: for job %s : %s" % (self.indexName, jobID, parameters))

        parametersListDict = [
            {"JobID": jobID, "Name": parName, "Value": parValue, "_id": str(parName) + str(parValue)}
            for parName, parValue in parameters
        ]

        result = self.bulk_index(self.indexName, data=parametersListDict, period=None, withTimeStamp=False)
        if not result["OK"]:
            self.log.error("ERROR: Couldn't insert data", result["Message"])
        return result

    def deleteJobParameters(self, jobID, paramList=None):
        """delete Job Parameters defined for jobID.
          Returns a dictionary with the Job Parameters.
          If paramList is empty - all the parameters for the job are removed

        :param self: self reference
        :param int jobID: Job ID
        :param list paramList: list of parameters to be returned (also a string is treated)

        :return: dict with all Job Parameter values
        """

        self.log.debug("JobDB.getParameters: Deleting Parameters %s for job %s" % (paramList, jobID))

        jobFilter = self._Q("term", JobID=jobID)

        if not paramList:
            s = self.dslSearch.query("bool", filter=jobFilter)
            s.delete()
            return S_OK()

        # the following should be equivalent to
        # {
        #   "query": {
        #     "bool": {
        #       "filter": [  # no scoring
        #         {"term": {"JobID": jobID}},  # term level query, does not pass through the analyzer
        #         {"term": {"Name": param}},  # term level query, does not pass through the analyzer
        #       ]
        #     }
        #   }
        # }

        if isinstance(paramList, six.string_types):
            paramList = paramList.replace(" ", "").split(",")

        for param in paramList:
            paramFilter = self._Q("term", Name=param)
            combinedFilter = jobFilter & paramFilter

            s = self.dslSearch.query("bool", filter=combinedFilter)
            s.delete()

        return S_OK()

    # TODO: Add query by value (e.g. query which values are in a certain pattern)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import copy
import unittest
import mock

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.Resources.Storage.StorageFactory import StorageFactory
from functools import reduce

gLogger.setLevel("DEBUG")
dict_cs = {
    "Resources": {
        "StorageElementBases": {
            "CERN-BASE-WITH-TWO-SAME-PLUGINS": {
                "BackendType": "Eos",
                "SEType": "T0D1",
                "AccessProtocol.1": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "PluginName": "GFAL2_SRM2",
                    "Protocol": "srm",
                    "Access": "remote",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
                "AccessProtocol.2": {
                    "Host": "eoslhcb.cern.ch",
                    "Port": 8443,
                    "PluginName": "GFAL2_SRM2",
                    "Protocol": "root",
                    "Access": "remote",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
            },
            # Pure abstract because no path or SpaceToken
            "CERN-ABSTRACT": {
                "BackendType": "Eos",
                "SEType": "T0D1",
                "AccessProtocol.1": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "PluginName": "GFAL2_SRM2",
                    "Protocol": "srm",
                    "Access": "remote",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
                "AccessProtocol.2": {
                    "Host": "eoslhcb.cern.ch",
                    "PluginName": "GFAL2_XROOT",
                    "Protocol": "root",
                    "Access": "remote",
                },
            },
            "CERN-BASE": {
                "BackendType": "Eos",
                "SEType": "T0D1",
                "AccessProtocol.1": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "PluginName": "GFAL2_SRM2",
                    "Protocol": "srm",
                    "Path": "/eos/lhcb/grid/prod",
                    "Access": "remote",
                    "SpaceToken": "LHCb-EOS",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
            },
        },
        "StorageElements": {
            # This SE must be in the section above. We kept the backward
            # compatibility for a while, but now it is time to remove it
            "CERN-BASE-WRONGLOCATION": {
                "BackendType": "Eos",
                "SEType": "T0D1",
                "AccessProtocol.1": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "PluginName": "GFAL2_SRM2",
                    "Protocol": "srm",
                    "Path": "/eos/lhcb/grid/prod",
                    "Access": "remote",
                    "SpaceToken": "LHCb-EOS",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
            },
            # This should not work anymore because
            # CERN-BASE-WRONGLOCATION is in StorageElements
            "CERN-WRONGLOCATION": {
                "BaseSE": "CERN-BASE-WRONGLOCATION",
                "AccessProtocol.1": {
                    "PluginName": "GFAL2_SRM2",
                    "Path": "/eos/lhcb/grid/prod",
                },
            },
            # Alias in StorageElements location should still work
            "CERN-WRONGLOCATION-ALIAS": {"Alias": "CERN-BASE-WRONGLOCATION"},
            "CERN-SIMPLE": {
                "BackendType": "Eos",
                "SEType": "T0D1",
                "RemoteAccessProtocol": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "PluginName": "GFAL2_SRM2",
                    "Protocol": "srm",
                    "Path": "/eos/lhcb/grid/prod",
                    "Access": "remote",
                    "SpaceToken": "LHCb-EOS",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
                "LocalAccessProtocol": {
                    "Host": "eoslhcb.cern.ch",
                    "PluginName": "File",
                    "Protocol": "file",
                    "Path": "/eos/lhcb/grid/prod",
                    "Access": "local",
                    "SpaceToken": "LHCb-EOS",
                },
            },
            # Just inherit, overwrite a SpaceToken and Path and add an option
            "CERN-USER": {
                "BaseSE": "CERN-BASE",
                "PledgedSpace": 205,
                "AccessProtocol.1": {
                    "PluginName": "GFAL2_SRM2",
                    "Path": "/eos/lhcb/grid/user",
                    "SpaceToken": "LHCb_USER",
                },
            },
            "CERN-DST": {
                "BaseSE": "CERN-BASE",
                "AccessProtocol.1": {
                    "PluginName": "GFAL2_SRM2",
                    "Path": "/eos/lhcb/grid/prod",
                },
            },
            # Does not redefine anything, it is just to give another name
            "CERN-NO-DEF": {
                "BaseSE": "CERN-BASE",
            },
            # Uses the plugin name of the Base SE
            "CERN-NO-PLUGIN-NAME": {
                "BaseSE": "CERN-BASE",
                "AccessProtocol.1": {
                    "Path": "/eos/lhcb/grid/user",
                },
            },
            # Redefines the plugin name in the protocol section
            "CERN-BAD-PLUGIN-NAME": {
                "BaseSE": "CERN-BASE",
                "AccessProtocol.1": {
                    "PluginName": "AnotherPluginName",
                    "Path": "/eos/lhcb/grid/prod",
                    "Access": "local",
                },
            },
            # Gives 2 protocol sections with the same plugin name
            "CERN-REDEFINE-PLUGIN-NAME": {
                "BaseSE": "CERN-BASE",
                "AccessProtocol.OtherName": {
                    "PluginName": "GFAL2_SRM2",
                    "Path": "/eos/lhcb/grid/other",
                    "Access": "remote",
                },
            },
            # The plugin name of the GFAL2_SRM2 section should be GFAL2_SRM2
            "CERN-USE-PLUGIN-AS-PROTOCOL-NAME": {
                "BaseSE": "CERN-BASE",
                "GFAL2_SRM2": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "Protocol": "srm",
                    "Path": "/eos/lhcb/grid/user",
                    "Access": "remote",
                    "SpaceToken": "LHCb-EOS",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
            },
            # Plugin name should be GFAL2_XROOT here
            "CERN-USE-PLUGIN-AS-PROTOCOL-NAME-WITH-PLUGIN-NAME": {
                "BaseSE": "CERN-BASE",
                "GFAL2_SRM2": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "Protocol": "srm",
                    "Path": "/eos/lhcb/grid/user",
                    "Access": "remote",
                    "SpaceToken": "LHCb-EOS",
                    "PluginName": "GFAL2_XROOT",
                    "WSUrl": "/srm/v2/server?SFN:",
                },
            },
            # Defines two same plugins in two protocol sections
            "CERN-CHILD-INHERIT-FROM-BASE-WITH-TWO-SAME-PLUGINS": {
                "BaseSE": "CERN-BASE-WITH-TWO-SAME-PLUGINS",
                "AccessProtocol.1": {
                    "Path": "/eos/lhcb/grid/user",
                },
            },
            # More because add an extra protocol compared to the parent
            "CERN-MORE": {
                "BaseSE": "CERN-BASE",
                "AccessProtocol.1": {
                    "PluginName": "GFAL2_SRM2",
                    "Path": "/eos/lhcb/grid/user",
                },
                "AccessProtocol.More": {
                    "Host": "srm-eoslhcb.cern.ch",
                    "Port": 8443,
                    "PluginName": "Extra",
                    "Protocol": "srm",
                    "Path": "/eos/lhcb/grid/prod",
                    "Access": "remote",
                    "SpaceToken": "LHCb-EOS",
                },
            },
            # Inherits from ABSTRACT
            "CERN-CHILD": {
                "BaseSE": "CERN-ABSTRACT",
                "AccessProtocol.1": {
                    "PluginName": "GFAL2_SRM2",
                    "Path": "/eos/lhcb/grid/user",
                    "SpaceToken": "LHCb_USER",
                },
                "AccessProtocol.2": {
                    "PluginName": "GFAL2_XROOT",
                    "Path": "/eos/lhcb/grid/xrootuser",
                },
            },
        },
    }
}


class fake_gConfig(object):
    @staticmethod
    def crawlCS(path):
        # How nice :-)
        # split the path and recursively dig down the dict_cs dictionary
        return reduce(lambda d, e: d.get(e, {}), path.strip("/").split("/"), dict_cs)

    def getValue(self, path, defaultValue=""):
        if "StorageElements" not in path and "StorageElementBases" not in path:
            return gConfig.getValue(path, defaultValue)
        csValue = self.crawlCS(path)
        if not csValue:
            csValue = defaultValue
        return csValue

    def getOptionsDict(self, path):
        """Mock the getOptionsDict call of gConfig
        It reads from dict_cs
        """
        if "StorageElements" not in path and "StorageElementBases" not in path:
            return gConfig.getOptionsDict(path)
        csSection = self.crawlCS(path)
        if not csSection:
            return S_ERROR("Not a valid section")
        options = dict((opt, val) for opt, val in csSection.items() if not isinstance(val, dict))
        return S_OK(options)

    def getOptions(self, path):
        """Mock the getOptions call of gConfig
        It reads from dict_cs
        """
        if "StorageElements" not in path and "StorageElementBases" not in path:
            return gConfig.getOptions(path)
        csSection = self.crawlCS(path)
        if not csSection:
            return S_ERROR("Not a valid section")
        options = [opt for opt, val in csSection.items() if not isinstance(val, dict)]
        return S_OK(options)

    def getSections(self, path):
        """Mock the getOptions call of gConfig
        It reads from dict_cs
        """
        if "StorageElements" not in path and "StorageElementBases" not in path:
            return gConfig.getSections(path)
        csSection = self.crawlCS(path)
        if not csSection:
            return S_ERROR("Not a valid section")
        sections = [opt for opt, val in csSection.items() if isinstance(val, dict)]
        return S_OK(sections)


def mock_StorageFactory__generateStorageObject(*args, **kwargs):
    """Don't really load the plugin, just create an object"""
    return S_OK(object())


def mock_resourceStatus_getElementStatus(seName, elementType="StorageElement"):
    """We shut up RSS"""
    return S_OK({seName: {}})


sf_gConfig = fake_gConfig()
mandatoryProtocolOptions = {
    "Access": "",
    "Host": "",
    "Path": "",
    "Port": "",
    "Protocol": "",
    "SpaceToken": "",
    "WSUrl": "",
}


@mock.patch("DIRAC.Resources.Storage.StorageFactory.gConfig", new=sf_gConfig)
@mock.patch(
    "DIRAC.Resources.Storage.StorageFactory.StorageFactory._StorageFactory__generateStorageObject",
    side_effect=mock_StorageFactory__generateStorageObject,
)
@mock.patch(
    "DIRAC.ResourceStatusSystem.Client.ResourceStatus.ResourceStatus.getElementStatus",
    side_effect=mock_resourceStatus_getElementStatus,
)
class StorageFactoryStandaloneTestCase(unittest.TestCase):
    """Base class for the StorageFactory test cases"""

    def test_standalone(self, _sf_generateStorageObject, _rss_getSEStatus):
        """Test loading a storage element with everything defined in itself.
        It should have two storage plugins
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-SIMPLE")

        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["LocalPlugins"], ["File"])
        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2"])

        allProtocols = []
        for protocol in ["RemoteAccessProtocol", "LocalAccessProtocol"]:
            protocolDef = copy.copy(mandatoryProtocolOptions)
            protocolDef.update(fake_gConfig.crawlCS("/Resources/StorageElements/CERN-SIMPLE/%s" % protocol))
            allProtocols.append(protocolDef)

        self.assertEqual(len(storages["ProtocolOptions"]), len(allProtocols))
        self.assertEqual(len(storages["StorageObjects"]), len(allProtocols))

        self.assertListEqual(
            sorted(allProtocols, key=lambda x: x["Host"]), sorted(storages["ProtocolOptions"], key=lambda x: x["Host"])
        )
        self.assertDictEqual(storages["StorageOptions"], {"BackendType": "Eos", "SEType": "T0D1"})


@mock.patch("DIRAC.Resources.Storage.StorageFactory.gConfig", new=sf_gConfig)
@mock.patch(
    "DIRAC.Resources.Storage.StorageFactory.StorageFactory._StorageFactory__generateStorageObject",
    side_effect=mock_StorageFactory__generateStorageObject,
)
@mock.patch(
    "DIRAC.ResourceStatusSystem.Client.ResourceStatus.ResourceStatus.getElementStatus",
    side_effect=mock_resourceStatus_getElementStatus,
)
class StorageFactorySimpleInheritance(unittest.TestCase):
    """In this class we perform simple inheritance test, with only one
    protocol and no extra protocol definition
    """

    def test_simple_inheritance_overwrite(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-USER that inherits from CERN-BASE,
        add a storage option, redefine the path and the space token
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-USER")

        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2"])

        # There should be a single protocol
        self.assertEqual(len(storages["ProtocolOptions"]), 1)
        # There should be one storage object
        self.assertEqual(len(storages["StorageObjects"]), 1)

        protocolDetail = storages["ProtocolOptions"][0]
        # These are the values we expect
        self.assertEqual(protocolDetail["Access"], "remote")
        self.assertEqual(protocolDetail["Host"], "srm-eoslhcb.cern.ch")
        self.assertEqual(protocolDetail["Path"], "/eos/lhcb/grid/user")
        self.assertEqual(protocolDetail["PluginName"], "GFAL2_SRM2")
        self.assertEqual(protocolDetail["Port"], 8443)
        self.assertEqual(protocolDetail["Protocol"], "srm")
        self.assertEqual(protocolDetail["SpaceToken"], "LHCb_USER")
        self.assertEqual(protocolDetail["WSUrl"], "/srm/v2/server?SFN:")

        self.assertDictEqual(
            storages["StorageOptions"],
            {"BackendType": "Eos", "SEType": "T0D1", "PledgedSpace": 205, "BaseSE": "CERN-BASE"},
        )

    def test_simple_inheritance(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-DST that inherits from CERN-BASE,
        and redefine the same value for Path and PluginName
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-DST")

        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2"])

        # There should be a single protocol
        self.assertEqual(len(storages["ProtocolOptions"]), 1)
        # There should be one storage object
        self.assertEqual(len(storages["StorageObjects"]), 1)

        protocolDetail = storages["ProtocolOptions"][0]
        # These are the values we expect
        self.assertEqual(protocolDetail["Access"], "remote")
        self.assertEqual(protocolDetail["Host"], "srm-eoslhcb.cern.ch")
        self.assertEqual(protocolDetail["Path"], "/eos/lhcb/grid/prod")
        self.assertEqual(protocolDetail["PluginName"], "GFAL2_SRM2")
        self.assertEqual(protocolDetail["Port"], 8443)
        self.assertEqual(protocolDetail["Protocol"], "srm")
        self.assertEqual(protocolDetail["SpaceToken"], "LHCb-EOS")
        self.assertEqual(protocolDetail["WSUrl"], "/srm/v2/server?SFN:")

        self.assertDictEqual(
            storages["StorageOptions"], {"BackendType": "Eos", "SEType": "T0D1", "BaseSE": "CERN-BASE"}
        )

    def test_pure_inheritance(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-NO-DEF that inherits from CERN-BASE,
        but does not redefine ANYTHING. We expect it to be just like the parent
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-NO-DEF")

        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2"])

        # There should be a single protocol
        self.assertEqual(len(storages["ProtocolOptions"]), 1)
        # There should be one storage object
        self.assertEqual(len(storages["StorageObjects"]), 1)

        protocolDetail = storages["ProtocolOptions"][0]
        # These are the values we expect
        self.assertEqual(protocolDetail["Access"], "remote")
        self.assertEqual(protocolDetail["Host"], "srm-eoslhcb.cern.ch")
        self.assertEqual(protocolDetail["Path"], "/eos/lhcb/grid/prod")
        self.assertEqual(protocolDetail["PluginName"], "GFAL2_SRM2")
        self.assertEqual(protocolDetail["Port"], 8443)
        self.assertEqual(protocolDetail["Protocol"], "srm")
        self.assertEqual(protocolDetail["SpaceToken"], "LHCb-EOS")
        self.assertEqual(protocolDetail["WSUrl"], "/srm/v2/server?SFN:")

        self.assertDictEqual(
            storages["StorageOptions"], {"BackendType": "Eos", "SEType": "T0D1", "BaseSE": "CERN-BASE"}
        )


@mock.patch("DIRAC.Resources.Storage.StorageFactory.gConfig", new=sf_gConfig)
@mock.patch(
    "DIRAC.Resources.Storage.StorageFactory.StorageFactory._StorageFactory__generateStorageObject",
    side_effect=mock_StorageFactory__generateStorageObject,
)
@mock.patch(
    "DIRAC.ResourceStatusSystem.Client.ResourceStatus.ResourceStatus.getElementStatus",
    side_effect=mock_resourceStatus_getElementStatus,
)
class StorageFactoryWeirdDefinition(unittest.TestCase):
    """In this class, we test very specific cases to highlight inheritance of the StorageElement"""

    def test_no_plugin_name(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-NO-PLUGIN-NAME that inherits from CERN-BASE,
        and redefine the same protocol but with no PluginName
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-NO-PLUGIN-NAME")

        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2"])

        expectedProtocols = [
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/user",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            }
        ]

        self.assertListEqual(storages["ProtocolOptions"], expectedProtocols)

    def test_bad_plugin_name(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-BAD-PLUGIN-NAME that inherits from CERN-BASE,
        and redefine the same protocol but with a different PluginName.
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-BAD-PLUGIN-NAME")
        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], [])
        self.assertListEqual(storages["LocalPlugins"], ["AnotherPluginName"])

        expectedProtocols = [
            {
                "Access": "local",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/prod",
                "PluginName": "AnotherPluginName",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            }
        ]

        self.assertListEqual(storages["ProtocolOptions"], expectedProtocols)

    def test_redefine_plugin_name(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-REDEFINE-PLUGIN-NAME that inherits from CERN-BASE,
        and uses the same Plugin with a different section.
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-REDEFINE-PLUGIN-NAME")
        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2", "GFAL2_SRM2"])

        expectedProtocols = [
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/prod",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            },
            {
                "Access": "remote",
                "Host": "",
                "Path": "/eos/lhcb/grid/other",
                "PluginName": "GFAL2_SRM2",
                "Port": "",
                "Protocol": "",
                "SpaceToken": "",
                "WSUrl": "",
            },
        ]

        self.assertListEqual(storages["ProtocolOptions"], expectedProtocols)

    def test_use_plugin_as_protocol_name(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-USE-PLUGIN-AS-PROTOCOL-NAME that inherits from CERN-BASE,
        and uses a protocol named as a plugin name, the plugin name is not present.
        """
        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-USE-PLUGIN-AS-PROTOCOL-NAME")
        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2", "GFAL2_SRM2"])

        expectedProtocols = [
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/prod",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            },
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/user",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            },
        ]

        self.assertListEqual(storages["ProtocolOptions"], expectedProtocols)

    def test_use_plugin_as_protocol_name_with_plugin_name(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-USE-PLUGIN-AS-PROTOCOL-NAME that inherits from CERN-BASE,
        and uses a protocol named as a plugin name, the plugin name is also present.
        """
        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-USE-PLUGIN-AS-PROTOCOL-NAME-WITH-PLUGIN-NAME")
        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2", "GFAL2_XROOT"])

        expectedProtocols = [
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/prod",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            },
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/user",
                "PluginName": "GFAL2_XROOT",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            },
        ]

        self.assertListEqual(storages["ProtocolOptions"], expectedProtocols)

    def test_more_protocol(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-MORE that inherits from CERN-BASE,
        and adds an extra protocol
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-MORE")
        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertSetEqual(set(storages["RemotePlugins"]), set(["Extra", "GFAL2_SRM2"]))

        expectedProtocols = [
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/prod",
                "PluginName": "Extra",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "",
            },
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/user",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb-EOS",
                "WSUrl": "/srm/v2/server?SFN:",
            },
        ]

        self.assertListEqual(
            sorted(storages["ProtocolOptions"], key=lambda x: x["PluginName"]),
            expectedProtocols,
        )

    def test_child_inherit_from_base_with_two_same_plugins(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-CHILD-INHERIT-FROM-BASE-WITH-TWO-SAME-PLUGINS that inherits
        from CERN-BASE-WITH-TWO-SAME-PLUGINS, using two identical plugin names in two sections.
        """
        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-CHILD-INHERIT-FROM-BASE-WITH-TWO-SAME-PLUGINS")
        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2", "GFAL2_SRM2"])

        expectedProtocols = [
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/user",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "",
                "WSUrl": "/srm/v2/server?SFN:",
            },
            {
                "Access": "remote",
                "Host": "eoslhcb.cern.ch",
                "Path": "",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "root",
                "SpaceToken": "",
                "WSUrl": "/srm/v2/server?SFN:",
            },
        ]

        self.assertListEqual(storages["ProtocolOptions"], expectedProtocols)

    def test_baseSE_in_SEDefinition(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, a storage inherits from a baseSE which is declared in the
        StorageElements section instead of the BaseStorageElements section.
        It used to be possible, but we remove this compatibility layer.
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-WRONGLOCATION")

        self.assertFalse(storages["OK"], storages)

    def test_aliasSE_in_SEDefinition(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, a storage aliases a baseSE which is declared in the
        StorageElements section. That should remain possible
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-WRONGLOCATION-ALIAS")

        self.assertTrue(storages["OK"], storages)


@mock.patch("DIRAC.Resources.Storage.StorageFactory.gConfig", new=sf_gConfig)
@mock.patch(
    "DIRAC.Resources.Storage.StorageFactory.StorageFactory._StorageFactory__generateStorageObject",
    side_effect=mock_StorageFactory__generateStorageObject,
)
@mock.patch(
    "DIRAC.ResourceStatusSystem.Client.ResourceStatus.ResourceStatus.getElementStatus",
    side_effect=mock_resourceStatus_getElementStatus,
)
class StorageFactoryPureAbstract(unittest.TestCase):
    """In this class, we test pure abstract inheritance"""

    def test_pure_abstract(self, _sf_generateStorageObject, _rss_getSEStatus):
        """In this test, we load a storage element CERN-CHILD that inherits from CERN-ABSTRACT.
        CERN-ABSTRACT has two uncomplete protocols, and CERN-CHILD defines them
        """

        sf = StorageFactory(vo="lhcb")
        storages = sf.getStorages("CERN-CHILD")
        self.assertTrue(storages["OK"], storages)
        storages = storages["Value"]

        self.assertListEqual(storages["RemotePlugins"], ["GFAL2_SRM2", "GFAL2_XROOT"])

        expectedProtocols = [
            {
                "Access": "remote",
                "Host": "srm-eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/user",
                "PluginName": "GFAL2_SRM2",
                "Port": 8443,
                "Protocol": "srm",
                "SpaceToken": "LHCb_USER",
                "WSUrl": "/srm/v2/server?SFN:",
            },
            {
                "Access": "remote",
                "Host": "eoslhcb.cern.ch",
                "Path": "/eos/lhcb/grid/xrootuser",
                "PluginName": "GFAL2_XROOT",
                "Port": "",
                "Protocol": "root",
                "SpaceToken": "",
                "WSUrl": "",
            },
        ]

        self.assertListEqual(storages["ProtocolOptions"], expectedProtocols)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(StorageFactoryStandaloneTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(StorageFactorySimpleInheritance))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(StorageFactoryWeirdDefinition))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(StorageFactoryPureAbstract))
    testResult = unittest.TextTestRunner(verbosity=3).run(suite)



""" Palette is a tool to generate colors for various Graphs plots and legends

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import hashlib

from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus

job_status_palette = {
    JobStatus.RECEIVED: "#D9E7F8",
    JobStatus.CHECKING: "#FAFAFA",
    JobStatus.STAGING: "#6190CD",
    JobStatus.WAITING: "#004EFF",
    JobStatus.MATCHED: "#FEF7AA",
    JobStatus.RUNNING: "#FDEE65",
    JobStatus.COMPLETING: "#FFAF55",
    JobStatus.STALLED: "#BC5757",
    JobStatus.COMPLETED: "#00FF21",
    JobStatus.DONE: "#238802",
    JobStatus.FAILED: "#FF0000",
    JobStatus.KILLED: "#111111",
}

job_minor_status_palette = {
    JobMinorStatus.APP_ERRORS: "#BC2133",
    JobMinorStatus.EXCEPTION_DURING_EXEC: "#AA240C",
    JobMinorStatus.EXEC_COMPLETE: "#338B39",
    JobMinorStatus.ILLEGAL_JOB_JDL: "#D96C00",
    JobMinorStatus.INPUT_NOT_AVAILABLE: "#2822A6",
    JobMinorStatus.INPUT_DATA_RESOLUTION: "#FFBE94",
    JobMinorStatus.DOWNLOADING_INPUT_SANDBOX: "#586CFF",
    JobMinorStatus.INPUT_CONTAINS_SLASHES: "#AB7800",
    JobMinorStatus.INPUT_INCORRECT: "#6812D6",
    JobMinorStatus.JOB_WRAPPER_INITIALIZATION: "#FFFFCC",
    JobMinorStatus.JOB_EXCEEDED_WALL_CLOCK: "#FF33CC",
    JobMinorStatus.JOB_INSUFFICIENT_DISK: "#33FFCC",
    JobMinorStatus.JOB_EXCEEDED_CPU: "#AABBCC",
    "No Ancestors Found For Input Data": "#BDA544",
    JobMinorStatus.NO_CANDIDATE_SITE_FOUND: "#E2FFBC",
    JobMinorStatus.PENDING_REQUESTS: "#52FF4F",
    JobMinorStatus.RECEIVED_KILL_SIGNAL: "#FF312F",
    JobMinorStatus.WATCHDOG_STALLED: "#FFCC99",
}

miscelaneous_pallette = {"Others": "#666666", "NoLabels": "#0025AD", "Total": "#00FFDC", "Default": "#FDEE65"}

country_palette = {
    "France": "#73C6BC",
    "UK": "#DCAF8A",
    "Spain": "#C2B0E1",
    "Netherlands": "#A9BF8E",
    "Germany": "#800000",
    "Russia": "#00514A",
    "Italy": "#004F00",
    "Switzerland": "#433B00",
    "Poland": "#528220",
    "Hungary": "#825CE2",
    "Portugal": "#009182",
    "Turkey": "#B85D00",
}


class Palette(object):
    def __init__(self, palette={}, colors=[]):

        self.palette = country_palette
        self.palette.update(job_status_palette)
        self.palette.update(miscelaneous_pallette)
        self.palette.update(job_minor_status_palette)

    def setPalette(self, palette):
        self.palette = palette

    def setColor(self, label, color):
        self.palette[label] = color

    def addPalette(self, palette):
        self.palette.update(palette)

    def getColor(self, label):

        if label in self.palette:
            return self.palette[label]
        else:
            return self.generateColor(label)

    def generateColor(self, label):

        myMD5 = hashlib.md5()
        myMD5.update(label.encode())
        hexstring = myMD5.hexdigest()
        color = "#" + hexstring[:6]
        return color



"""   The Job Scheduling Executor takes the information gained from all previous
      optimizers and makes a scheduling decision for the jobs.

      Subsequent to this jobs are added into a Task Queue and pilot agents can be submitted.

      All issues preventing the successful resolution of a site candidate are discovered
      here where all information is available.

      This Executor will fail affected jobs meaningfully.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id: $"

import random

from DIRAC import S_OK, S_ERROR, gConfig

from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.Core.Utilities.Time import fromString, toEpoch
from DIRAC.Core.Security import Properties
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus
from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient, getFilesToStage
from DIRAC.WorkloadManagementSystem.Executor.Base.OptimizerExecutor import OptimizerExecutor
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.Client import JobStatus


class JobScheduling(OptimizerExecutor):
    """
    The specific Optimizer must provide the following methods:
    - optimizeJob() - the main method called for each job
    and it can provide:
    - initializeOptimizer() before each execution cycle
    """

    @classmethod
    def initializeOptimizer(cls):
        """Initialization of the optimizer."""
        cls.siteClient = SiteStatus()
        cls.__jobDB = JobDB()
        return S_OK()

    def optimizeJob(self, jid, jobState):
        """1. Banned sites are removed from the destination list.
        2. Get input files
        3. Production jobs are sent directly to TQ
        4. Check if staging is necessary
        """
        # Reschedule delay
        result = jobState.getAttributes(["RescheduleCounter", "RescheduleTime", "ApplicationStatus"])
        if not result["OK"]:
            return result
        attDict = result["Value"]
        try:
            reschedules = int(attDict["RescheduleCounter"])
        except (ValueError, KeyError):
            return S_ERROR("RescheduleCounter has to be an integer")
        if reschedules != 0:
            delays = self.ex_getOption("RescheduleDelays", [60, 180, 300, 600])
            delay = delays[min(reschedules, len(delays) - 1)]
            waited = toEpoch() - toEpoch(fromString(attDict["RescheduleTime"]))
            if waited < delay:
                return self.__holdJob(jobState, "On Hold: after rescheduling %s" % reschedules, delay)

        # Get the job manifest for the later checks
        result = jobState.getManifest()
        if not result["OK"]:
            self.jobLog.error("Could not retrieve job manifest", result["Message"])
            return result
        jobManifest = result["Value"]

        # Get site requirements
        result = self.__getSitesRequired(jobManifest)
        if not result["OK"]:
            return result
        userSites, userBannedSites = result["Value"]

        # Get job type
        result = jobState.getAttribute("JobType")
        if not result["OK"]:
            self.jobLog.error("Could not retrieve job type", result["Message"])
            return result
        jobType = result["Value"]

        # Get banned sites from DIRAC
        result = self.siteClient.getSites("Banned")
        if not result["OK"]:
            self.jobLog.error("Cannot retrieve banned sites", result["Message"])
            return result
        wmsBannedSites = result["Value"]

        # If the user has selected any site, filter them and hold the job if not able to run
        if userSites:
            if jobType not in self.ex_getOption("ExcludedOnHoldJobTypes", []):

                result = self.siteClient.getUsableSites(userSites)
                if not result["OK"]:
                    self.jobLog.error(
                        "Problem checking userSites for tuple of active/banned/invalid sites", result["Message"]
                    )
                    return result
                usableSites = set(result["Value"])
                bannedSites = []
                invalidSites = []
                for site in userSites:
                    if site in wmsBannedSites:
                        bannedSites.append(site)
                    elif site not in usableSites:
                        invalidSites.append(site)

                if invalidSites:
                    self.jobLog.debug("Invalid site(s) requested: %s" % ",".join(invalidSites))
                    if not self.ex_getOption("AllowInvalidSites", True):
                        return self.__holdJob(jobState, "Requested site(s) %s are invalid" % ",".join(invalidSites))
                if bannedSites:
                    self.jobLog.debug("Banned site(s) %s ignored" % ",".join(bannedSites))
                    if not usableSites:
                        return self.__holdJob(jobState, "Requested site(s) %s are inactive" % ",".join(bannedSites))

                if not usableSites:
                    return self.__holdJob(jobState, "No requested site(s) are active/valid")
                userSites = list(usableSites)

        checkPlatform = self.ex_getOption("CheckPlatform", False)
        jobPlatform = jobManifest.getOption("Platform", None)
        # First check that the platform is valid (in OSCompatibility list)
        if checkPlatform and jobPlatform:
            result = gConfig.getOptionsDict("/Resources/Computing/OSCompatibility")
            if not result["OK"]:
                self.jobLog.error("Unable to get OSCompatibility list", result["Message"])
                return result
            allPlatforms = result["Value"]
            if jobPlatform not in allPlatforms:
                self.jobLog.error("Platform not supported", jobPlatform)
                return S_ERROR("Platform is not supported")

        # Filter the userSites by the platform selection (if there is one)
        if checkPlatform and userSites:
            if jobPlatform:
                result = self.__filterByPlatform(jobPlatform, userSites)
                if not result["OK"]:
                    self.jobLog.error("Failed to filter job sites by platform", result["Message"])
                    return result
                userSites = result["Value"]
                if not userSites:
                    # No sites left after filtering -> Invalid platform/sites combination
                    self.jobLog.error("No selected sites match platform", jobPlatform)
                    return S_ERROR("No selected sites match platform '%s'" % jobPlatform)

        # Check if there is input data
        result = jobState.getInputData()
        if not result["OK"]:
            self.jobLog.error("Failed to get input data from JobDB", result["Message"])
            return result

        if not result["Value"]:
            # No input data? Just send to TQ
            return self.__sendToTQ(jobState, jobManifest, userSites, userBannedSites)

        self.jobLog.verbose("Has an input data requirement")
        inputData = result["Value"]

        # ===================================================================================
        # Production jobs are sent to TQ, but first we have to verify if staging is necessary
        # ===================================================================================
        if jobType in Operations().getValue("Transformations/DataProcessing", []):
            self.jobLog.info("Production job: sending to TQ, but first checking if staging is requested")

            res = getFilesToStage(
                inputData,
                jobState=jobState,
                checkOnlyTapeSEs=self.ex_getOption("CheckOnlyTapeSEs", True),
                jobLog=self.jobLog,
            )

            if not res["OK"]:
                return self.__holdJob(jobState, res["Message"])
            if res["Value"]["absentLFNs"]:
                # Some files do not exist at all... set the job Failed
                # Reverse errors
                reasons = {}
                for lfn, reason in res["Value"]["absentLFNs"].items():
                    reasons.setdefault(reason, []).append(lfn)
                for reason, lfns in reasons.items():
                    # Some files are missing in the FC or in SEs, fail the job
                    self.jobLog.error(reason, ",".join(lfns))
                error = ",".join(reasons)
                return S_ERROR(error)

            if res["Value"]["failedLFNs"]:
                return self.__holdJob(jobState, "Couldn't get storage metadata of some files")
            stageLFNs = res["Value"]["offlineLFNs"]
            if stageLFNs:
                res = self.__checkStageAllowed(jobState)
                if not res["OK"]:
                    return res
                if not res["Value"]:
                    return S_ERROR("Stage not allowed")
                self.__requestStaging(jobState, stageLFNs)
                return S_OK()
            else:
                # No staging required
                onlineSites = res["Value"]["onlineSites"]
                if onlineSites:
                    # Set the online site(s) first
                    userSites = set(userSites)
                    onlineSites &= userSites
                    userSites = list(onlineSites) + list(userSites - onlineSites)
                return self.__sendToTQ(jobState, jobManifest, userSites, userBannedSites, onlineSites=onlineSites)

        # ===================================================
        # From now on we know it's a user job with input data
        # ===================================================

        idAgent = self.ex_getOption("InputDataAgent", "InputData")
        result = self.retrieveOptimizerParam(idAgent)
        if not result["OK"]:
            self.jobLog.error("Could not retrieve input data info", result["Message"])
            return result
        opData = result["Value"]

        if "SiteCandidates" not in opData:
            return S_ERROR("No possible site candidates")

        # Filter input data sites with user requirement
        siteCandidates = list(opData["SiteCandidates"])
        self.jobLog.info("Site candidates are", siteCandidates)

        if userSites:
            siteCandidates = list(set(siteCandidates) & set(userSites))

        siteCandidates = self._applySiteFilter(siteCandidates, banned=userBannedSites)
        if not siteCandidates:
            return S_ERROR("Impossible InputData * Site requirements")

        idSites = {}
        for site in siteCandidates:
            idSites[site] = opData["SiteCandidates"][site]

        # Check if sites have correct count of disk+tape replicas
        numData = len(inputData)
        errorSites = set()
        for site in idSites:
            if numData != idSites[site]["disk"] + idSites[site]["tape"]:
                self.jobLog.error("Site candidate does not have all the input data", "(%s)" % site)
                errorSites.add(site)
        for site in errorSites:
            idSites.pop(site)
        if not idSites:
            return S_ERROR("Site candidates do not have all the input data")

        # Check if staging is required
        stageRequired, siteCandidates = self.__resolveStaging(inputData, idSites)
        if not siteCandidates:
            return S_ERROR("No destination sites available")

        # Is any site active?
        stageSites = self._applySiteFilter(siteCandidates, banned=wmsBannedSites)
        if not stageSites:
            return self.__holdJob(jobState, "Sites %s are inactive or banned" % ", ".join(siteCandidates))

        # If no staging is required send to TQ
        if not stageRequired:
            # Use siteCandidates and not stageSites because active and banned sites
            # will be taken into account on matching time
            return self.__sendToTQ(jobState, jobManifest, siteCandidates, userBannedSites)

        # Check if the user is allowed to stage
        if self.ex_getOption("RestrictDataStage", False):
            res = self.__checkStageAllowed(jobState)
            if not res["OK"]:
                return res
            if not res["Value"]:
                return S_ERROR("Stage not allowed")

        # Get stageSites[0] because it has already been randomized and it's as good as any in stageSites
        stageSite = stageSites[0]
        self.jobLog.verbose(" Staging site will be", stageSite)
        stageData = idSites[stageSite]
        # Set as if everything has already been staged
        stageData["disk"] += stageData["tape"]
        stageData["tape"] = 0
        # Set the site info back to the original dict to save afterwards
        opData["SiteCandidates"][stageSite] = stageData

        stageRequest = self.__preRequestStaging(jobManifest, stageSite, opData)
        if not stageRequest["OK"]:
            return stageRequest
        stageLFNs = stageRequest["Value"]
        result = self.__requestStaging(jobState, stageLFNs)
        if not result["OK"]:
            return result
        stageLFNs = result["Value"]
        self.__updateSharedSESites(jobManifest, stageSite, stageLFNs, opData)
        # Save the optimizer data again
        self.jobLog.verbose("Updating Optimizer Info", ": %s for %s" % (idAgent, opData))
        result = self.storeOptimizerParam(idAgent, opData)
        if not result["OK"]:
            return result

        return self.__setJobSite(jobState, stageSites)

    def _applySiteFilter(self, sites, banned=False):
        """Filters out banned sites"""
        if not sites:
            return sites

        filtered = set(sites)
        if banned and isinstance(banned, (list, set, dict)):
            filtered -= set(banned)
        return list(filtered)

    def __holdJob(self, jobState, holdMsg, delay=0):
        if delay:
            self.freezeTask(delay)
        else:
            self.freezeTask(self.ex_getOption("HoldTime", 300))
        self.jobLog.info("On hold", holdMsg)
        return jobState.setAppStatus(holdMsg, source=self.ex_optimizerName())

    def __getSitesRequired(self, jobManifest):
        """Returns any candidate sites specified by the job or sites that have been
        banned and could affect the scheduling decision.
        """

        bannedSites = jobManifest.getOption("BannedSites", [])
        if not bannedSites:
            bannedSites = jobManifest.getOption("BannedSite", [])
        if bannedSites:
            self.jobLog.info("Banned sites", ", ".join(bannedSites))

        sites = jobManifest.getOption("Site", [])
        # TODO: Only accept known sites after removing crap like ANY set in the original manifest
        sites = [site for site in sites if site.strip().lower() not in ("any", "")]

        if sites:
            if len(sites) == 1:
                self.jobLog.info("Single chosen site", ": %s specified" % (sites[0]))
            else:
                self.jobLog.info("Multiple sites requested", ": %s" % ",".join(sites))
            sites = self._applySiteFilter(sites, banned=bannedSites)
            if not sites:
                return S_ERROR("Impossible site requirement")

        return S_OK((sites, bannedSites))

    def __filterByPlatform(self, jobPlatform, userSites):
        """Filters out sites that have no CE with a matching platform."""
        basePath = "/Resources/Sites"
        filteredSites = set()

        # FIXME: can use Resources().getSiteCEMapping()
        for site in userSites:
            if "." not in site:
                # Invalid site name: Doesn't contain a dot!
                self.jobLog.warn("Skipped invalid site name", site)
                continue
            grid = site.split(".")[0]
            sitePath = cfgPath(basePath, grid, site, "CEs")
            result = gConfig.getSections(sitePath)
            if not result["OK"]:
                self.jobLog.info("Failed to get CEs", "at site %s" % site)
                continue
            siteCEs = result["Value"]

            for CEName in siteCEs:
                CEPlatform = gConfig.getValue(cfgPath(sitePath, CEName, "OS"))
                if jobPlatform == CEPlatform:
                    # Site has a CE with a matchin platform
                    filteredSites.add(site)

        return S_OK(list(filteredSites))

    def _getTagsFromManifest(self, jobManifest):
        """helper method to add a list of tags to the TQ from the job manifest content"""

        # Generate Tags from specific requirements
        tagList = []

        # sorting out the number of processors
        nProcessors = 1
        maxProcessors = 1

        if "NumberOfProcessors" in jobManifest:  # this should be the exact number
            nProcessors = jobManifest.getOption("NumberOfProcessors", 0)

        else:  # is there a min? and in that case, is there a max?
            if "MinNumberOfProcessors" in jobManifest:
                nProcessors = jobManifest.getOption("MinNumberOfProcessors", 0)

                if "MaxNumberOfProcessors" in jobManifest:
                    maxProcessors = jobManifest.getOption("MaxNumberOfProcessors", 0)
                else:
                    maxProcessors = -1

        if nProcessors and nProcessors > 1:
            tagList.append("%dProcessors" % nProcessors)
            tagList.append("MultiProcessor")
        if maxProcessors == -1 or maxProcessors > 1:
            tagList.append("MultiProcessor")

        if "WholeNode" in jobManifest:
            if jobManifest.getOption("WholeNode", "").lower() in ["1", "yes", "true", "y"]:
                tagList.append("WholeNode")
                tagList.append("MultiProcessor")

        # sorting out the RAM (this should be probably coded ~same as number of processors)
        if "MaxRAM" in jobManifest:
            maxRAM = jobManifest.getOption("MaxRAM", 0)
            if maxRAM:
                tagList.append("%dGB" % maxRAM)

        # other tags? Just add them
        if "Tags" in jobManifest:
            tagList.extend(jobManifest.getOption("Tags", []))
        if "Tag" in jobManifest:
            tagList.extend(jobManifest.getOption("Tag", []))

        return tagList

    def __sendToTQ(self, jobState, jobManifest, sites, bannedSites, onlineSites=None):
        """This method sends jobs to the task queue agent and if candidate sites
        are defined, updates job JDL accordingly.
        """

        tagList = self._getTagsFromManifest(jobManifest)
        if tagList:
            jobManifest.setOption("Tags", ", ".join(tagList))

        reqSection = "JobRequirements"
        if reqSection in jobManifest:
            result = jobManifest.getSection(reqSection)
        else:
            result = jobManifest.createSection(reqSection)
        if not result["OK"]:
            self.jobLog.error("Cannot create jobManifest section", "(%s: %s)" % reqSection, result["Message"])
            return result
        reqCfg = result["Value"]

        if sites:
            reqCfg.setOption("Sites", ", ".join(sites))
        if bannedSites:
            reqCfg.setOption("BannedSites", ", ".join(bannedSites))

        # Job multivalue requirement keys are specified as singles in the job descriptions
        # but for backward compatibility can be also plurals
        for key in ("SubmitPools", "SubmitPool", "JobType", "GridRequiredCEs", "GridCE", "Tags"):
            reqKey = key
            if key == "JobType":
                reqKey = "JobTypes"
            elif key == "GridRequiredCEs" or key == "GridCE":  # Remove obsolete GridRequiredCEs
                reqKey = "GridCEs"
            elif key == "SubmitPools" or key == "SubmitPool":  # Check: one of these 2 is obsolete
                reqKey = "SubmitPools"
            if key in jobManifest:
                reqCfg.setOption(reqKey, ", ".join(jobManifest.getOption(key, [])))

        result = self.__setJobSite(jobState, sites, onlineSites=onlineSites)
        if not result["OK"]:
            return result

        self.jobLog.verbose("Done")
        return self.setNextOptimizer(jobState)

    def __resolveStaging(self, inputData, idSites):
        diskSites = []
        maxOnDisk = 0
        bestSites = []

        for site in idSites:
            nTape = idSites[site]["tape"]
            nDisk = idSites[site]["disk"]
            if nTape > 0:
                self.jobLog.debug("%s tape replicas on site %s" % (nTape, site))
            if nDisk > 0:
                self.jobLog.debug("%s disk replicas on site %s" % (nDisk, site))
                if nDisk == len(inputData):
                    diskSites.append(site)
            if nDisk > maxOnDisk:
                maxOnDisk = nDisk
                bestSites = [site]
            elif nDisk == maxOnDisk:
                bestSites.append(site)

        # If there are selected sites, those are disk only sites
        if diskSites:
            self.jobLog.verbose("No staging required")
            return (False, diskSites)

        self.jobLog.verbose("Staging required")
        if len(bestSites) > 1:
            random.shuffle(bestSites)
        return (True, bestSites)

    def __preRequestStaging(self, jobManifest, stageSite, opData):

        tapeSEs = []
        diskSEs = []
        vo = jobManifest.getOption("VirtualOrganization")
        inputDataPolicy = jobManifest.getOption("InputDataPolicy", "Protocol")
        connectionLevel = "DOWNLOAD" if "download" in inputDataPolicy.lower() else "PROTOCOL"
        # Allow staging from SEs accessible by protocol
        result = DMSHelpers(vo=vo).getSEsForSite(stageSite, connectionLevel=connectionLevel)
        if not result["OK"]:
            return S_ERROR("Could not determine SEs for site %s" % stageSite)
        siteSEs = result["Value"]

        for seName in siteSEs:
            se = StorageElement(seName, vo=vo)
            seStatus = se.getStatus()
            if not seStatus["OK"]:
                return seStatus
            seStatus = seStatus["Value"]
            if seStatus["Read"] and seStatus["TapeSE"]:
                tapeSEs.append(seName)
            if seStatus["Read"] and seStatus["DiskSE"]:
                diskSEs.append(seName)

        if not tapeSEs:
            return S_ERROR("No Local SEs for site %s" % stageSite)

        self.jobLog.debug("Tape SEs are %s" % (", ".join(tapeSEs)))

        # I swear this is horrible DM code it's not mine.
        # Eternity of hell to the inventor of the Value of Value of Success of...
        inputData = opData["Value"]["Value"]["Successful"]
        stageLFNs = {}
        lfnToStage = []
        for lfn in inputData:
            replicas = inputData[lfn]
            # Check SEs
            seStage = []
            for seName in replicas:
                if seName in diskSEs:
                    # This lfn is in disk. Skip it
                    seStage = []
                    break
                if seName not in tapeSEs:
                    # This lfn is not in this tape SE. Check next SE
                    continue
                seStage.append(seName)
            for seName in seStage:
                if seName not in stageLFNs:
                    stageLFNs[seName] = []
                stageLFNs[seName].append(lfn)
                if lfn not in lfnToStage:
                    lfnToStage.append(lfn)

        if not stageLFNs:
            return S_ERROR("Cannot find tape replicas")

        # Check if any LFN is in more than one SE
        # If that's the case, try to stage from the SE that has more LFNs to stage to group the request
        # 1.- Get the SEs ordered by ascending replicas
        sortedSEs = reversed(sorted([(len(stageLFNs[seName]), seName) for seName in stageLFNs]))
        for lfn in lfnToStage:
            found = False
            # 2.- Traverse the SEs
            for _stageCount, seName in sortedSEs:
                if lfn in stageLFNs[seName]:
                    # 3.- If first time found, just mark as found. Next time delete the replica from the request
                    if found:
                        stageLFNs[seName].remove(lfn)
                    else:
                        found = True
                # 4.-If empty SE, remove
                if not stageLFNs[seName]:
                    stageLFNs.pop(seName)

        return S_OK(stageLFNs)

    def __requestStaging(self, jobState, stageLFNs):
        """Actual request for staging LFNs through the StorageManagerClient"""
        self.jobLog.debug(
            "Stage request will be \n\t%s" % "\n\t".join(["%s:%s" % (lfn, stageLFNs[lfn]) for lfn in stageLFNs])
        )

        stagerClient = StorageManagerClient()
        result = jobState.setStatus(
            JobStatus.STAGING,
            self.ex_getOption("StagingMinorStatus", "Request To Be Sent"),
            appStatus="",
            source=self.ex_optimizerName(),
        )
        if not result["OK"]:
            return result

        result = stagerClient.setRequest(
            stageLFNs, "WorkloadManagement", "updateJobFromStager@WorkloadManagement/JobStateUpdate", int(jobState.jid)
        )
        if not result["OK"]:
            self.jobLog.error("Could not send stage request", ": %s" % result["Message"])
            return result

        rid = str(result["Value"])
        self.jobLog.info("Stage request sent", "(%s)" % rid)
        self.storeOptimizerParam("StageRequest", rid)

        result = jobState.setStatus(
            JobStatus.STAGING,
            self.ex_getOption("StagingMinorStatus", "Request Sent"),
            appStatus="",
            source=self.ex_optimizerName(),
        )
        if not result["OK"]:
            return result

        return S_OK(stageLFNs)

    def __updateSharedSESites(self, jobManifest, stageSite, stagedLFNs, opData):
        siteCandidates = opData["SiteCandidates"]

        seStatus = {}
        vo = jobManifest.getOption("VirtualOrganization")
        for siteName in siteCandidates:
            if siteName == stageSite:
                continue
            self.jobLog.debug("Checking %s for shared SEs" % siteName)
            siteData = siteCandidates[siteName]
            result = getSEsForSite(siteName)
            if not result["OK"]:
                continue
            closeSEs = result["Value"]
            diskSEs = []
            for seName in closeSEs:
                # If we don't have the SE status get it and store it
                if seName not in seStatus:
                    seStatus[seName] = StorageElement(seName, vo=vo).status()
                # get the SE status from mem and add it if its disk
                status = seStatus[seName]
                if status["Read"] and status["DiskSE"]:
                    diskSEs.append(seName)
            self.jobLog.debug("Disk SEs for %s are %s" % (siteName, ", ".join(diskSEs)))

            # Hell again to the dev of this crappy value of value of successful of ...
            lfnData = opData["Value"]["Value"]["Successful"]
            for seName in stagedLFNs:
                # If the SE is not close then skip it
                if seName not in closeSEs:
                    continue
                for lfn in stagedLFNs[seName]:
                    self.jobLog.debug("Checking %s for %s" % (seName, lfn))
                    # I'm pretty sure that this cannot happen :P
                    if lfn not in lfnData:
                        continue
                    # Check if it's already on disk at the site
                    onDisk = False
                    for siteSE in lfnData[lfn]:
                        if siteSE in diskSEs:
                            self.jobLog.verbose("lfn on disk", ": %s at %s" % (lfn, siteSE))
                            onDisk = True
                    # If not on disk, then update!
                    if not onDisk:
                        self.jobLog.verbose("Setting LFN to disk", "for %s" % seName)
                        siteData["disk"] += 1
                        siteData["tape"] -= 1

    def __setJobSite(self, jobState, siteList, onlineSites=None):
        """Set the site attribute"""
        if onlineSites is None:
            onlineSites = []
        numSites = len(siteList)
        if numSites == 0:
            self.jobLog.info("Any site is candidate")
            return jobState.setAttribute("Site", "ANY")
        elif numSites == 1:
            self.jobLog.info("Only 1 site is candidate", ": %s" % siteList[0])
            return jobState.setAttribute("Site", siteList[0])

        # If the job has input data, the online sites are hosting the data
        if len(onlineSites) == 1:
            siteName = "Group.%s" % ".".join(list(onlineSites)[0].split(".")[1:])
            self.jobLog.info("Group %s is candidate" % siteName)
        elif onlineSites:
            # More than one site with input
            siteName = "MultipleInput"
            self.jobLog.info("Several input sites are candidate", ": %s" % ",".join(onlineSites))
        else:
            # No input site reported (could be a user job)
            siteName = "Multiple"
            self.jobLog.info("Multiple sites are candidate")

        return jobState.setAttribute("Site", siteName)

    def __checkStageAllowed(self, jobState):
        """Check if the job credentials allow to stage date"""
        result = jobState.getAttribute("OwnerGroup")
        if not result["OK"]:
            self.jobLog.error("Cannot retrieve OwnerGroup from DB", ": %s" % result["Message"])
            return result
        group = result["Value"]
        return S_OK(Properties.STAGE_ALLOWED in Registry.getPropertiesForGroup(group))



"""
It used to create several plots
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import time
import copy

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Plotting import gDataCache
from DIRAC.Core.Utilities.Plotting.Plots import (
    generateNoDataPlot,
    generateTimedStackedBarPlot,
    generateQualityPlot,
    generateCumulativePlot,
    generatePiePlot,
    generateStackedLinePlot,
)

from DIRAC.MonitoringSystem.private.DBUtils import DBUtils

__RCSID__ = "$Id$"


class BasePlotter(DBUtils):

    _EA_THUMBNAIL = "thumbnail"
    _EA_WIDTH = "width"
    _EA_HEIGHT = "height"
    _EA_THB_WIDTH = "thbWidth"
    _EA_THB_HEIGHT = "thbHeight"
    _EA_PADDING = "figurePadding"
    _EA_TITLE = "plotTitle"

    _RATE_UNITS = {
        "time": (
            ("seconds / s", 1, 24),
            ("hours / s", 3600, 24),
            ("days / s", 86400, 15),
            ("weeks / s", 86400 * 7, 10),
            ("months / s", 86400 * 30, 12),
            ("years / s", 86400 * 365, 1),
        ),
        "cpupower": (("HS06", 1, 750), ("kHS06", 1000, 750), ("MHS06", 10 ** 6, 1)),
        "bytes": (
            ("MB / s", 10 ** 6, 1000),
            ("GB / s", 10 ** 9, 1000),
            ("TB / s", 10 ** 12, 1000),
            ("PB / s", 10 ** 15, 1),
        ),
        "jobs": (
            ("jobs / hour", 1 / 3600.0, 1000),
            ("kjobs / hour", (10 ** 3) / 3600.0, 1000),
            ("Mjobs / hour", (10 ** 6) / 3600.0, 1),
        ),
        "files": (
            ("files / hour", 1 / 3600.0, 1000),
            ("kfiles / hour", (10 ** 3) / 3600.0, 1000),
            ("Mfiles / hour", (10 ** 6) / 3600.0, 1),
        ),
    }

    _UNITS = {
        "time": (
            ("seconds", 1, 24),
            ("hours", 3600, 24),
            ("days", 86400, 15),
            ("weeks", 86400 * 7, 10),
            ("months", 86400 * 30, 12),
            ("years", 86400 * 365, 1),
        ),
        "cpupower": (
            ("HS06 hours", 3600, 24),
            ("HS06 days", 86400, 750),
            ("kHS06 days", 86400 * 1000, 750),
            ("MHS06 days", 86400 * 10 ** 6, 1),
        ),
        "bytes": (("MB", 10 ** 6, 1000), ("GB", 10 ** 9, 1000), ("TB", 10 ** 12, 1000), ("PB", 10 ** 15, 1)),
        "jobs": (("jobs", 1, 1000), ("kjobs", 10 ** 3, 1000), ("Mjobs", 10 ** 6, 1)),
        "files": (("files", 1, 1000), ("kfiles", 10 ** 3, 1000), ("Mfiles", 10 ** 6, 1)),
    }

    # To be defined in the derived classes
    _typeKeyFields = []
    _typeName = ""

    def __init__(self, db, setup, extraArgs=None):
        super(BasePlotter, self).__init__(db, setup)
        """ c'tor

    :param self: self reference
    """

        if isinstance(extraArgs, dict):
            self._extraArgs = extraArgs
        else:
            self._extraArgs = {}
        reportsRevMap = {}
        for attr in dir(self):
            if attr.startswith("_report"):
                if attr.endswith("Name"):
                    reportId = attr[7:-4]
                    reportName = getattr(self, attr)
                    reportsRevMap[reportId] = reportName
                else:
                    reportId = attr[7:]
                    if reportId not in reportsRevMap:
                        reportsRevMap[reportId] = reportId
        self.__reportNameMapping = {}
        for rId in reportsRevMap:
            self.__reportNameMapping[reportsRevMap[rId]] = rId

    def generate(self, reportRequest):
        """
        It retrives the data from the database and create the plot

        :param dict reportRequest: contains the plot attributes
        """
        reportHash = reportRequest["hash"]
        reportName = reportRequest["reportName"]
        if reportName in self.__reportNameMapping:
            reportRequest["reportName"] = self.__reportNameMapping[reportName]

        gLogger.info("Retrieving data for %s:%s" % (reportRequest["typeName"], reportRequest["reportName"]))
        sT = time.time()
        retVal = self.__retrieveReportData(reportRequest, reportHash)
        reportGenerationTime = time.time() - sT
        if not retVal["OK"]:
            return retVal
        if not reportRequest["generatePlot"]:
            return retVal
        reportData = retVal["Value"]
        gLogger.info("Plotting data for %s:%s" % (reportRequest["typeName"], reportRequest["reportName"]))
        sT = time.time()
        retVal = self.__generatePlotForReport(reportRequest, reportHash, reportData)
        plotGenerationTime = time.time() - sT
        gLogger.verbose(
            "Time for %s:%s - Report %.2f Plot %.2f (%.2f%% r/p)"
            % (
                reportRequest["typeName"],
                reportRequest["reportName"],
                reportGenerationTime,
                plotGenerationTime,
                ((reportGenerationTime * 100 / plotGenerationTime) if plotGenerationTime else 0.0),
            )
        )
        if not retVal["OK"]:
            return retVal
        plotDict = retVal["Value"]
        if "retrieveReportData" in reportRequest["extraArgs"] and reportRequest["extraArgs"]["retrieveReportData"]:
            plotDict["reportData"] = reportData
        return S_OK(plotDict)

    def plotsList(self):
        """
        It returns the list of available plots.
        """
        return sorted(self.__reportNameMapping)

    def __retrieveReportData(self, reportRequest, reportHash):
        """
        It uses the appropriate Plotter to retrieve the data from the database.

        :param dict reportRequest: the dictionary which contains the conditions used to create the plot
        :param str reportHash: it is the unique identifier used to cache a plot
        :return: dict S_OK/S_ERROR if the data found in the cache it returns from it otherwise it uses the cache.
        """
        funcName = "_report%s" % reportRequest["reportName"]
        if not hasattr(self, funcName):
            return S_ERROR("Report %s is not defined" % reportRequest["reportName"])
        else:
            funcObj = getattr(self, funcName)

        return gDataCache.getReportData(reportRequest=reportRequest, reportHash=reportHash, dataFunc=funcObj)

    def __generatePlotForReport(self, reportRequest, reportHash, reportData):
        """It creates the plot

        :param dict reportRequest: contains the plot attributes
        :param str reportHash: unique string which identify the plot
        :param dict repotData: contains the data used to generate the plot.
        """

        funcName = "_plot%s" % reportRequest["reportName"]
        try:
            funcObj = getattr(self, funcName)
        except Exception:
            return S_ERROR("Plot function for report %s is not defined" % reportRequest["reportName"])

        return gDataCache.getReportPlot(
            reportRequest=reportRequest, reportHash=reportHash, reportData=reportData, plotFunc=funcObj
        )

    def _getTimedData(self, startTime, endTime, selectField, preCondDict, metadataDict=None):
        """
        It retrieves the time series data from the ES database.

        :param int startTime: epoch time
        :param int endTime: epoch time
        :param str selectField: the value that we want to plot
        :param dict preCondDict: plot attributes
        :param dict metadataDict: extra arguments used to create the plot.

        """

        condDict = {}

        if metadataDict is None:
            metadataDict = {}

        grouping = preCondDict["grouping"][0]
        # Make safe selections
        for keyword in self._typeKeyFields:
            if keyword in preCondDict:
                condDict[keyword] = preCondDict[keyword]

        retVal = self._determineBucketSize(startTime, endTime)
        if not retVal["OK"]:
            return retVal
        interval, granularity = retVal["Value"]

        dynamicBucketing = metadataDict.get("DynamicBucketing", True)
        # by default we use dynamic bucketing
        if dynamicBucketing:
            retVal = self._retrieveBucketedData(
                self._typeName, startTime, endTime, interval, selectField, condDict, grouping, metadataDict
            )
        else:
            retVal = self._retrieveAggregatedData(
                self._typeName, startTime, endTime, interval, selectField, condDict, grouping, metadataDict
            )

        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]

        return S_OK((dataDict, granularity))

    def _getSummaryData(self, startTime, endTime, selectField, preCondDict, metadataDict=None):
        """
        It returns the data used to create the pie chart plot.

        :param int startTime: epoch time
        :param int endTime: epoch time
        :param str selectField: the value what we want to plot
        :param dict preCondDict: plot attributes
        :param dict metadataDict: extra arguments used to create the plot.
        """
        grouping = preCondDict["grouping"][0]
        condDict = {}
        # Make safe selections
        for keyword in self._typeKeyFields:
            if keyword in preCondDict:
                condDict[keyword] = preCondDict[keyword]

        retVal = self._determineBucketSize(startTime, endTime)
        if not retVal["OK"]:
            return retVal
        interval, _ = retVal["Value"]

        retVal = self._retrieveBucketedData(
            typeName=self._typeName,
            startTime=startTime,
            endTime=endTime,
            interval=interval,
            selectField=selectField,
            condDict=condDict,
            grouping=grouping,
            metadataDict=metadataDict,
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        return S_OK(dataDict)

    def _findSuitableRateUnit(self, dataDict, maxValue, unit):
        """
        Returns the suitable unit for a given dataset.
        """
        return self._findUnitMagic(dataDict, maxValue, unit, self._RATE_UNITS)

    def _findSuitableUnit(self, dataDict, maxValue, unit):
        """
        Returns the suitable unit for a given dataset.
        """
        return self._findUnitMagic(dataDict, maxValue, unit, self._UNITS)

    def _findUnitMagic(self, reportDataDict, maxValue, unit, selectedUnits):
        """
        Returns the suitable unit for a given dataset.
        """
        if unit not in selectedUnits:
            raise AttributeError("%s is not a known rate unit" % unit)
        baseUnitData = selectedUnits[unit][0]
        if self._extraArgs.get("staticUnits"):
            unitData = selectedUnits[unit][0]
        else:
            unitList = selectedUnits[unit]
            unitIndex = -1
            for _, unitDivFactor, unitThreshold in unitList:
                unitIndex += 1
                if maxValue / unitDivFactor < unitThreshold:
                    break
            unitData = selectedUnits[unit][unitIndex]
        # Apply divFactor to all units
        graphDataDict, maxValue = self._divideByFactor(copy.deepcopy(reportDataDict), unitData[1])
        if unitData == baseUnitData:
            reportDataDict = graphDataDict
        else:
            reportDataDict, dummyMaxValue = self._divideByFactor(reportDataDict, baseUnitData[1])
        return reportDataDict, graphDataDict, maxValue, unitData[0]

    def __checkPlotMetadata(self, metadata):
        """It check the plot metadata arguments

        :param dict metadata: contains the plot metadata
        """
        if self._extraArgs.get(self._EA_WIDTH):
            try:
                metadata[self._EA_WIDTH] = min(1600, max(200, int(self._extraArgs[self._EA_WIDTH])))
            except Exception:
                pass
        if self._EA_HEIGHT in self._extraArgs and self._extraArgs[self._EA_HEIGHT]:
            try:
                metadata[self._EA_HEIGHT] = min(1600, max(200, int(self._extraArgs[self._EA_HEIGHT])))
            except Exception:
                pass
        if self._extraArgs.get(self._EA_TITLE):
            metadata["title"] = self._extraArgs[self._EA_TITLE]

    def __checkThumbnailMetadata(self, metadata):
        """checks the plot thumbnail data

        :param dict metadata: contains the thumbnail data
        """
        if self._EA_THUMBNAIL in self._extraArgs and self._extraArgs[self._EA_THUMBNAIL]:
            thbMD = dict(metadata)
            thbMD["legend"] = False
            if self._EA_THB_HEIGHT in self._extraArgs:
                thbMD[self._EA_HEIGHT] = self._extraArgs[self._EA_THB_HEIGHT]
            else:
                thbMD[self._EA_HEIGHT] = 125
            if self._EA_THB_WIDTH in self._extraArgs:
                thbMD[self._EA_WIDTH] = self._extraArgs[self._EA_THB_WIDTH]
            else:
                thbMD[self._EA_WIDTH] = 200
            thbMD[self._EA_PADDING] = 20
            for key in ("title", "ylabel", "xlabel"):
                if key in thbMD:
                    del thbMD[key]
            return thbMD
        return False

    def __plotData(self, filename, dataDict, metadata, funcToPlot):
        """It create the plot.

        :param str filename: the name of the file which contains the plot
        :param dict dataDict: data used to crate the plot
        :param dict metadata: plot metadata
        :param object funcToPlot: the method which create the plot using the appropriate method.
        """
        self.__checkPlotMetadata(metadata)
        if not dataDict:
            funcToPlot = generateNoDataPlot
        plotFileName = "%s.png" % filename

        finalResult = funcToPlot(fileName=plotFileName, data=dataDict, metadata=metadata)
        if not finalResult["OK"]:
            return finalResult
        thbMD = self.__checkThumbnailMetadata(metadata)
        if not thbMD:
            return S_OK({"plot": True, "thumbnail": False})
        thbFilename = "%s.thb.png" % filename
        retVal = funcToPlot(thbFilename, dataDict, thbMD)
        if not retVal["OK"]:
            return retVal
        return S_OK({"plot": True, "thumbnail": True})

    def _generateTimedStackedBarPlot(self, filename, dataDict, metadata):
        """
        it creates a bar plot
        """
        return self.__plotData(filename, dataDict, metadata, generateTimedStackedBarPlot)

    def _generateQualityPlot(self, filename, dataDict, metadata):
        """
        it create a quality plot
        """
        return self.__plotData(filename, dataDict, metadata, generateQualityPlot)

    def _generateCumulativePlot(self, filename, dataDict, metadata):
        """
        It creates a cumulative plot
        """
        return self.__plotData(filename, dataDict, metadata, generateCumulativePlot)

    def _generatePiePlot(self, filename, dataDict, metadata):
        """
        It creates a pie chart plot
        """
        return self.__plotData(filename, dataDict, metadata, generatePiePlot)

    def _generateStackedLinePlot(self, filename, dataDict, metadata):
        """
        It create a stacked lien plot
        """
        return self.__plotData(filename, dataDict, metadata, generateStackedLinePlot)

    def _fillWithZero(self, granularity, startEpoch, endEpoch, dataDict):
        """
        Fill with zeros missing buckets
          - dataDict = { 'key' : { time1 : value,  time2 : value... }, 'key2'.. }
        """
        startBucketEpoch = startEpoch - startEpoch % granularity
        for key in dataDict:
            currentDict = dataDict[key]
            for timeEpoch in range(int(startBucketEpoch), int(endEpoch), granularity):
                if timeEpoch not in currentDict:
                    currentDict[timeEpoch] = 0
        return dataDict
